# Copyright (c) 2012 NetApp, Inc.  All rights reserved.
# Copyright (c) 2014 Ben Swartzlander.  All rights reserved.
# Copyright (c) 2014 Navneet Singh.  All rights reserved.
# Copyright (c) 2014 Clinton Knight.  All rights reserved.
# Copyright (c) 2014 Alex Meade.  All rights reserved.
# Copyright (c) 2014 Andrew Kerr.  All rights reserved.
# Copyright (c) 2014 Jeff Applewhite.  All rights reserved.
# Copyright (c) 2015 Tom Barron.  All rights reserved.
# Copyright (c) 2015 Dustin Schoenbrun. All rights reserved.
# Copyright (c) 2016 Chuck Fouts. All rights reserved.
# Copyright (c) 2016 Mike Rooney. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Volume driver library for NetApp 7/C-mode block storage systems.
"""

import copy
import math
import sys
import uuid

from oslo_log import log as logging
from oslo_log import versionutils
from oslo_utils import excutils
from oslo_utils import units
import six

from cinder import exception
from cinder.i18n import _
from cinder.volume.drivers.netapp.dataontap.client import api as netapp_api
from cinder.volume.drivers.netapp.dataontap.utils import loopingcalls
from cinder.volume.drivers.netapp import options as na_opts
from cinder.volume.drivers.netapp import utils as na_utils
from cinder.volume import volume_utils
from cinder.zonemanager import utils as fczm_utils

LOG = logging.getLogger(__name__)


class NetAppLun(object):
    """Represents a LUN on NetApp storage."""

    def __init__(self, handle, name, size, metadata_dict):
        self.handle = handle
        self.name = name
        self.size = size
        self.metadata = metadata_dict or {}

    def get_metadata_property(self, prop):
        """Get the metadata property of a LUN."""
        if prop in self.metadata:
            return self.metadata[prop]
        name = self.name
        LOG.debug("No metadata property %(prop)s defined for the LUN %(name)s",
                  {'prop': prop, 'name': name})

    def __str__(self, *args, **kwargs):
        return 'NetApp LUN [handle:%s, name:%s, size:%s, metadata:%s]' % (
               self.handle, self.name, self.size, self.metadata)


@six.add_metaclass(volume_utils.TraceWrapperMetaclass)
class NetAppBlockStorageLibrary(object):
    """NetApp block storage library for Data ONTAP."""

    # do not increment this as it may be used in volume type definitions
    VERSION = "1.0.0"
    REQUIRED_FLAGS = ['netapp_login', 'netapp_password',
                      'netapp_server_hostname']
    ALLOWED_LUN_OS_TYPES = ['linux', 'aix', 'hpux', 'image', 'windows',
                            'windows_2008', 'windows_gpt', 'solaris',
                            'solaris_efi', 'netware', 'openvms', 'hyper_v']
    ALLOWED_IGROUP_HOST_TYPES = ['linux', 'aix', 'hpux', 'windows', 'solaris',
                                 'netware', 'default', 'vmware', 'openvms',
                                 'xen', 'hyper_v']
    DEFAULT_LUN_OS = 'linux'
    DEFAULT_HOST_TYPE = 'linux'
    DEFAULT_FILTER_FUNCTION = 'capabilities.utilization < 70'
    DEFAULT_GOODNESS_FUNCTION = '100 - capabilities.utilization'

    def __init__(self, driver_name, driver_protocol, **kwargs):

        na_utils.validate_instantiation(**kwargs)

        self.driver_name = driver_name
        self.driver_protocol = driver_protocol
        self.zapi_client = None
        self._stats = {}
        self.lun_table = {}
        self.lun_ostype = None
        self.host_type = None
        self.lun_space_reservation = 'true'
        self.lookup_service = fczm_utils.create_lookup_service()
        self.app_version = kwargs.get("app_version", "unknown")
        self.host = kwargs.get('host')
        self.backend_name = self.host.split('@')[1]

        self.configuration = kwargs['configuration']
        self.configuration.append_config_values(na_opts.netapp_connection_opts)
        self.configuration.append_config_values(na_opts.netapp_basicauth_opts)
        self.configuration.append_config_values(na_opts.netapp_transport_opts)
        self.configuration.append_config_values(
            na_opts.netapp_provisioning_opts)
        self.configuration.append_config_values(na_opts.netapp_san_opts)
        self.max_over_subscription_ratio = (
            volume_utils.get_max_over_subscription_ratio(
                self.configuration.max_over_subscription_ratio,
                supports_auto=True))
        self.reserved_percentage = self._get_reserved_percentage()
        self.loopingcalls = loopingcalls.LoopingCalls()

    def _get_reserved_percentage(self):
        # If the legacy config option if it is set to the default
        # value, use the more general configuration option.
        if self.configuration.netapp_size_multiplier == (
                na_opts.NETAPP_SIZE_MULTIPLIER_DEFAULT):
            return self.configuration.reserved_percentage

        # If the legacy config option has a non-default value,
        # honor it for one release.  Note that the "size multiplier"
        # actually acted as a divisor in the code and didn't apply
        # to the file size (as the help message for this option suggest),
        # but rather to total and free size for the pool.
        divisor = self.configuration.netapp_size_multiplier
        reserved_ratio = round(1 - (1 / divisor), 2)
        reserved_percentage = 100 * int(reserved_ratio)
        msg = ('The "netapp_size_multiplier" configuration option is '
               'deprecated and will be removed in the Mitaka release. '
               'Please set "reserved_percentage = %d" instead.') % (
                   reserved_percentage)
        versionutils.report_deprecated_feature(LOG, msg)
        return reserved_percentage

    def do_setup(self, context):
        na_utils.check_flags(self.REQUIRED_FLAGS, self.configuration)
        self.lun_ostype = (self.configuration.netapp_lun_ostype
                           or self.DEFAULT_LUN_OS)
        self.host_type = (self.configuration.netapp_host_type
                          or self.DEFAULT_HOST_TYPE)
        if self.configuration.netapp_lun_space_reservation == 'enabled':
            self.lun_space_reservation = 'true'
        else:
            self.lun_space_reservation = 'false'

    def check_for_setup_error(self):
        """Check that the driver is working and can communicate.

        Discovers the LUNs on the NetApp server.
        """
        if self.lun_ostype not in self.ALLOWED_LUN_OS_TYPES:
            msg = _("Invalid value for NetApp configuration"
                    " option netapp_lun_ostype.")
            LOG.error(msg)
            raise na_utils.NetAppDriverException(msg)
        if self.host_type not in self.ALLOWED_IGROUP_HOST_TYPES:
            msg = _("Invalid value for NetApp configuration"
                    " option netapp_host_type.")
            LOG.error(msg)
            raise na_utils.NetAppDriverException(msg)
        lun_list = self.zapi_client.get_lun_list()
        self._extract_and_populate_luns(lun_list)
        LOG.debug("Success getting list of LUNs from server.")
        self.loopingcalls.start_tasks()

    def _add_looping_tasks(self):
        """Add tasks that need to be executed at a fixed interval.

        Inheriting class overrides and then explicitly calls this method.
        """

        # Add the task that deletes snapshots marked for deletion.
        self.loopingcalls.add_task(
            self._delete_snapshots_marked_for_deletion,
            loopingcalls.ONE_MINUTE,
            loopingcalls.ONE_MINUTE)

        # Add the task that logs EMS messages
        self.loopingcalls.add_task(
            self._handle_ems_logging,
            loopingcalls.ONE_HOUR)

    def _delete_snapshots_marked_for_deletion(self):
        snapshots = self.zapi_client.get_snapshots_marked_for_deletion()
        for snapshot in snapshots:
            self.zapi_client.delete_snapshot(
                snapshot['volume_name'], snapshot['name'])

    def _handle_ems_logging(self):
        """Log autosupport messages."""
        raise NotImplementedError()

    def get_pool(self, volume):
        """Return pool name where volume resides.

        :param volume: The volume hosted by the driver.
        :return: Name of the pool where given volume is hosted.
        """
        name = volume['name']
        metadata = self._get_lun_attr(name, 'metadata') or dict()
        return metadata.get('Volume', None)

    def create_volume(self, volume):
        """Driver entry point for creating a new volume (Data ONTAP LUN)."""

        LOG.debug('create_volume on %s', volume['host'])

        # get Data ONTAP volume name as pool name
        pool_name = volume_utils.extract_host(volume['host'], level='pool')

        if pool_name is None:
            msg = _("Pool is not available in the volume host field.")
            raise exception.InvalidHost(reason=msg)

        extra_specs = na_utils.get_volume_extra_specs(volume)

        lun_name = volume['name']

        size = int(volume['size']) * units.Gi

        metadata = {'OsType': self.lun_ostype,
                    'SpaceReserved': self.lun_space_reservation,
                    'Path': '/vol/%s/%s' % (pool_name, lun_name)}

        qos_policy_group_info = self._setup_qos_for_volume(volume, extra_specs)
        qos_policy_group_name = (
            na_utils.get_qos_policy_group_name_from_info(
                qos_policy_group_info))
        qos_policy_group_is_adaptive = (volume_utils.is_boolean_str(
            extra_specs.get('netapp:qos_policy_group_is_adaptive')) or
            na_utils.is_qos_policy_group_spec_adaptive(
                qos_policy_group_info))

        try:
            self._create_lun(pool_name, lun_name, size, metadata,
                             qos_policy_group_name,
                             qos_policy_group_is_adaptive)
        except Exception:
            LOG.exception("Exception creating LUN %(name)s in pool %(pool)s.",
                          {'name': lun_name, 'pool': pool_name})
            self._mark_qos_policy_group_for_deletion(qos_policy_group_info)
            msg = _("Volume %s could not be created.")
            raise exception.VolumeBackendAPIException(data=msg % (
                volume['name']))
        LOG.debug('Created LUN with name %(name)s and QoS info %(qos)s',
                  {'name': lun_name, 'qos': qos_policy_group_info})

        metadata['Path'] = '/vol/%s/%s' % (pool_name, lun_name)
        metadata['Volume'] = pool_name
        metadata['Qtree'] = None

        handle = self._create_lun_handle(metadata)
        self._add_lun_to_table(NetAppLun(handle, lun_name, size, metadata))

        model_update = self._get_volume_model_update(volume)

        return model_update

    def _setup_qos_for_volume(self, volume, extra_specs):
        return None

    def _get_volume_model_update(self, volume):
        """Provide any updates necessary for a volume being created/managed."""
        raise NotImplementedError

    def _mark_qos_policy_group_for_deletion(self, qos_policy_group_info):
        return

    def delete_volume(self, volume):
        """Driver entry point for destroying existing volumes."""
        self._delete_lun(volume['name'])

    def _delete_lun(self, lun_name):
        """Helper method to delete LUN backing a volume or snapshot."""

        metadata = self._get_lun_attr(lun_name, 'metadata')
        if metadata:
            try:
                self.zapi_client.destroy_lun(metadata['Path'])
            except netapp_api.NaApiError as e:
                if e.code == netapp_api.EOBJECTNOTFOUND:
                    LOG.warning("Failure deleting LUN %(name)s. %(message)s",
                                {'name': lun_name, 'message': e})
                else:
                    error_message = (_('A NetApp Api Error occurred: %s') % e)
                    raise na_utils.NetAppDriverException(error_message)
            self.lun_table.pop(lun_name)
        else:
            LOG.warning("No entry in LUN table for volume/snapshot"
                        " %(name)s.", {'name': lun_name})

    def ensure_export(self, context, volume):
        """Driver entry point to get the export info for an existing volume."""
        handle = self._get_lun_attr(volume['name'], 'handle')
        return {'provider_location': handle}

    def create_export(self, context, volume):
        """Driver entry point to get the export info for a new volume."""
        handle = self._get_lun_attr(volume['name'], 'handle')
        return {'provider_location': handle}

    def remove_export(self, context, volume):
        """Driver entry point to remove an export for a volume.

        Since exporting is idempotent in this driver, we have nothing
        to do for unexporting.
        """

        pass

    def create_snapshot(self, snapshot):
        """Driver entry point for creating a snapshot.

        This driver implements snapshots by using efficient single-file
        (LUN) cloning.
        """
        self._create_snapshot(snapshot)

    def _create_snapshot(self, snapshot):
        vol_name = snapshot['volume_name']
        snapshot_name = snapshot['name']
        lun = self._get_lun_from_table(vol_name)
        self._clone_lun(lun.name, snapshot_name, space_reserved='false',
                        is_snapshot=True)

    def delete_snapshot(self, snapshot):
        """Driver entry point for deleting a snapshot."""
        self._delete_lun(snapshot['name'])
        LOG.debug("Snapshot %s deletion successful", snapshot['name'])

    def create_volume_from_snapshot(self, volume, snapshot):
        source = {'name': snapshot['name'], 'size': snapshot['volume_size']}
        return self._clone_source_to_destination(source, volume)

    def create_cloned_volume(self, volume, src_vref):
        src_lun = self._get_lun_from_table(src_vref['name'])
        source = {'name': src_lun.name, 'size': src_vref['size']}
        return self._clone_source_to_destination(source, volume)

    def _clone_source_to_destination(self, source, destination_volume):
        source_size = source['size']
        destination_size = destination_volume['size']

        source_name = source['name']
        destination_name = destination_volume['name']

        extra_specs = na_utils.get_volume_extra_specs(destination_volume)

        qos_policy_group_info = self._setup_qos_for_volume(
            destination_volume, extra_specs)
        qos_policy_group_name = (
            na_utils.get_qos_policy_group_name_from_info(
                qos_policy_group_info))
        qos_policy_group_is_adaptive = (volume_utils.is_boolean_str(
            extra_specs.get('netapp:qos_policy_group_is_adaptive')) or
            na_utils.is_qos_policy_group_spec_adaptive(
                qos_policy_group_info))

        try:
            self._clone_lun(
                source_name, destination_name,
                space_reserved=self.lun_space_reservation,
                qos_policy_group_name=qos_policy_group_name,
                qos_policy_group_is_adaptive=qos_policy_group_is_adaptive)

            if destination_size != source_size:

                try:
                    self._extend_volume(destination_volume,
                                        destination_size,
                                        qos_policy_group_name)
                except Exception:
                    with excutils.save_and_reraise_exception():
                        LOG.error("Resizing %s failed. Cleaning volume.",
                                  destination_volume['id'])
                        self.delete_volume(destination_volume)

            return self._get_volume_model_update(destination_volume)

        except Exception:
            LOG.exception("Exception cloning volume %(name)s from source "
                          "volume %(source)s.",
                          {'name': destination_name, 'source': source_name})

            self._mark_qos_policy_group_for_deletion(qos_policy_group_info)

            msg = _("Volume %s could not be created from source volume.")
            raise exception.VolumeBackendAPIException(
                data=msg % destination_name)

    def _create_lun(self, volume_name, lun_name, size,
                    metadata, qos_policy_group_name=None):
        """Creates a LUN, handling Data ONTAP differences as needed."""
        raise NotImplementedError()

    def _create_lun_handle(self, metadata):
        """Returns LUN handle based on filer type."""
        raise NotImplementedError()

    def _extract_lun_info(self, lun):
        """Extracts the LUNs from API and populates the LUN table."""

        meta_dict = self._create_lun_meta(lun)
        path = lun.get_child_content('path')
        (_rest, _splitter, name) = path.rpartition('/')
        handle = self._create_lun_handle(meta_dict)
        size = lun.get_child_content('size')
        return NetAppLun(handle, name, size, meta_dict)

    def _extract_and_populate_luns(self, api_luns):
        """Extracts the LUNs from API and populates the LUN table."""

        for lun in api_luns:
            discovered_lun = self._extract_lun_info(lun)
            self._add_lun_to_table(discovered_lun)

    def _map_lun(self, name, initiator_list, initiator_type, lun_id=None):
        """Maps LUN to the initiator(s) and returns LUN ID assigned."""
        metadata = self._get_lun_attr(name, 'metadata')
        path = metadata['Path']
        igroup_name, ig_host_os, ig_type = self._get_or_create_igroup(
            initiator_list, initiator_type, self.host_type)
        if ig_host_os != self.host_type:
            LOG.warning("LUN misalignment may occur for current"
                        " initiator group %(ig_nm)s) with host OS type"
                        " %(ig_os)s. Please configure initiator group"
                        " manually according to the type of the"
                        " host OS.",
                        {'ig_nm': igroup_name, 'ig_os': ig_host_os})
        try:
            return self.zapi_client.map_lun(path, igroup_name, lun_id=lun_id)
        except netapp_api.NaApiError:
            exc_info = sys.exc_info()
            (_igroup, lun_id) = self._find_mapped_lun_igroup(path,
                                                             initiator_list)
            if lun_id is not None:
                return lun_id
            else:
                six.reraise(*exc_info)

    def _unmap_lun(self, path, initiator_list):
        """Unmaps a LUN from given initiator."""

        if len(initiator_list) != 0:
            lun_unmap_list = []
            (igroup_name, _) = self._find_mapped_lun_igroup(
                path, initiator_list)
            lun_unmap_list.append((path, igroup_name))
        else:
            lun_maps = self.zapi_client.get_lun_map(path)
            lun_unmap_list = [(path, lun_m['initiator-group'])
                              for lun_m in lun_maps]

        for _path, _igroup_name in lun_unmap_list:
            self.zapi_client.unmap_lun(_path, _igroup_name)

    def _find_mapped_lun_igroup(self, path, initiator_list):
        """Find an igroup for a LUN mapped to the given initiator(s)."""
        raise NotImplementedError()

    def _has_luns_mapped_to_initiators(self, initiator_list):
        """Checks whether any LUNs are mapped to the given initiator(s)."""
        return self.zapi_client.has_luns_mapped_to_initiators(initiator_list)

    def _get_or_create_igroup(self, initiator_list, initiator_group_type,
                              host_os_type):
        """Checks for an igroup for a set of one or more initiators.

        Creates igroup if not already present with given host os type,
        igroup type and adds initiators.
        """
        igroups = self.zapi_client.get_igroup_by_initiators(initiator_list)
        igroup_name = None

        if igroups:
            igroup = igroups[0]
            igroup_name = igroup['initiator-group-name']
            host_os_type = igroup['initiator-group-os-type']
            initiator_group_type = igroup['initiator-group-type']

        if not igroup_name:
            igroup_name = self._create_igroup_add_initiators(
                initiator_group_type, host_os_type, initiator_list)
        return igroup_name, host_os_type, initiator_group_type

    def _create_igroup_add_initiators(self, initiator_group_type,
                                      host_os_type, initiator_list):
        """Creates igroup and adds initiators."""
        igroup_name = na_utils.OPENSTACK_PREFIX + six.text_type(uuid.uuid4())
        self.zapi_client.create_igroup(igroup_name, initiator_group_type,
                                       host_os_type)
        for initiator in initiator_list:
            self.zapi_client.add_igroup_initiator(igroup_name, initiator)
        return igroup_name

    def _delete_lun_from_table(self, name):
        """Deletes LUN from cache table."""
        if self.lun_table.get(name, None):
            self.lun_table.pop(name)

    def _add_lun_to_table(self, lun):
        """Adds LUN to cache table."""
        if not isinstance(lun, NetAppLun):
            msg = _("Object is not a NetApp LUN.")
            raise exception.VolumeBackendAPIException(data=msg)
        self.lun_table[lun.name] = lun

    def _get_lun_from_table(self, name):
        """Gets LUN from cache table.

        Refreshes cache if LUN not found in cache.
        """
        lun = self.lun_table.get(name)
        if lun is None:
            lun_list = self.zapi_client.get_lun_list()
            self._extract_and_populate_luns(lun_list)
            lun = self.lun_table.get(name)
            if lun is None:
                raise exception.VolumeNotFound(volume_id=name)
        return lun

    def _clone_lun(self, name, new_name, space_reserved='true',
                   qos_policy_group_name=None, src_block=0, dest_block=0,
                   block_count=0, source_snapshot=None, is_snapshot=False):
        """Clone LUN with the given name to the new name."""
        raise NotImplementedError()

    def _get_lun_attr(self, name, attr):
        """Get the LUN attribute if found else None."""
        try:
            attr = getattr(self._get_lun_from_table(name), attr)
            return attr
        except exception.VolumeNotFound as e:
            LOG.error("Message: %s", e.msg)
        except Exception as e:
            LOG.error("Error getting LUN attribute. Exception: %s", e)
        return None

    def _create_lun_meta(self, lun):
        raise NotImplementedError()

    def _get_fc_target_wwpns(self, include_partner=True):
        raise NotImplementedError()

    def get_volume_stats(self, refresh=False, filter_function=None,
                         goodness_function=None):
        """Get volume stats.

        If 'refresh' is True, update the stats first.
        """

        if refresh:
            self._update_volume_stats(filter_function=filter_function,
                                      goodness_function=goodness_function)
        return self._stats

    def _update_volume_stats(self, filter_function=None,
                             goodness_function=None):
        raise NotImplementedError()

    def get_default_filter_function(self):
        """Get the default filter_function string."""
        return self.DEFAULT_FILTER_FUNCTION

    def get_default_goodness_function(self):
        """Get the default goodness_function string."""
        return self.DEFAULT_GOODNESS_FUNCTION

    def extend_volume(self, volume, new_size):
        """Driver entry point to increase the size of a volume."""

        extra_specs = na_utils.get_volume_extra_specs(volume)

        # Create volume copy with new size for size-dependent QOS specs
        volume_copy = copy.copy(volume)
        volume_copy['size'] = new_size

        qos_policy_group_info = self._setup_qos_for_volume(volume_copy,
                                                           extra_specs)
        qos_policy_group_name = (
            na_utils.get_qos_policy_group_name_from_info(
                qos_policy_group_info))

        try:
            self._extend_volume(volume, new_size, qos_policy_group_name)
        except Exception:
            with excutils.save_and_reraise_exception():
                # If anything went wrong, revert QoS settings
                self._setup_qos_for_volume(volume, extra_specs)

    def _extend_volume(self, volume, new_size, qos_policy_group_name):
        """Extend an existing volume to the new size."""
        name = volume['name']
        lun = self._get_lun_from_table(name)
        path = lun.metadata['Path']
        curr_size_bytes = six.text_type(lun.size)
        new_size_bytes = six.text_type(int(new_size) * units.Gi)
        # Reused by clone scenarios.
        # Hence comparing the stored size.
        if curr_size_bytes == new_size_bytes:
            LOG.info("No need to extend volume %s"
                     " as it is already the requested new size.", name)
            return

        ontap_version = self.zapi_client.get_ontap_version(cached=True)

        if ontap_version >= '9.5':
            self.zapi_client.do_direct_resize(path, new_size_bytes)
        else:
            lun_geometry = self.zapi_client.get_lun_geometry(path)
            if (lun_geometry and int(lun_geometry.get("max_resize", 0)) >=
                    int(new_size_bytes)):
                self.zapi_client.do_direct_resize(path, new_size_bytes)
            else:
                if volume['attach_status'] != 'detached':
                    msg = _('Volume %(vol_id)s cannot be resized from '
                            '%(old_size)s to %(new_size)s, because would '
                            'exceed its max geometry %(max_geo)s while '
                            'not being detached.')
                    raise exception.VolumeBackendAPIException(data=msg % {
                        'vol_id': name,
                        'old_size': curr_size_bytes,
                        'new_size': new_size_bytes,
                        'max_geo': lun_geometry.get("max_resize")})
                self._do_sub_clone_resize(
                    path, new_size_bytes,
                    qos_policy_group_name=qos_policy_group_name)
        self.lun_table[name].size = new_size_bytes

    def _get_vol_option(self, volume_name, option_name):
        """Get the value for the volume option."""
        value = None
        options = self.zapi_client.get_volume_options(volume_name)
        for opt in options:
            if opt.get_child_content('name') == option_name:
                value = opt.get_child_content('value')
                break
        return value

    def _do_sub_clone_resize(self, lun_path, new_size_bytes,
                             qos_policy_group_name=None):
        """Resize a LUN beyond its original geometry using sub-LUN cloning.

        Clones the block ranges, swaps the LUNs, and deletes the source LUN.
        """
        seg = lun_path.split("/")
        LOG.info("Resizing LUN %s using clone operation.", seg[-1])
        lun_name = seg[-1]
        vol_name = seg[2]
        lun = self._get_lun_from_table(lun_name)
        metadata = lun.metadata

        compression = self._get_vol_option(vol_name, 'compression')
        if compression == "on":
            msg = _('%s cannot be resized using clone operation'
                    ' as it is hosted on compressed volume')
            raise exception.VolumeBackendAPIException(data=msg % lun_name)

        block_count = self._get_lun_block_count(lun_path)
        if block_count == 0:
            msg = _('%s cannot be resized using clone operation'
                    ' as it contains no blocks.')
            raise exception.VolumeBackendAPIException(data=msg % lun_name)

        new_lun_name = 'new-%s' % lun_name
        self.zapi_client.create_lun(
            vol_name, new_lun_name, new_size_bytes, metadata,
            qos_policy_group_name=qos_policy_group_name)
        try:
            self._clone_lun(lun_name, new_lun_name, block_count=block_count)
            self._post_sub_clone_resize(lun_path)
        except Exception:
            with excutils.save_and_reraise_exception():
                new_lun_path = '/vol/%s/%s' % (vol_name, new_lun_name)
                self.zapi_client.destroy_lun(new_lun_path)

    def _post_sub_clone_resize(self, path):
        """Try post sub clone resize in a transactional manner."""
        st_tm_mv, st_nw_mv, st_del_old = None, None, None
        seg = path.split("/")
        LOG.info("Post clone resize LUN %s", seg[-1])
        new_lun = 'new-%s' % (seg[-1])
        tmp_lun = 'tmp-%s' % (seg[-1])
        tmp_path = "/vol/%s/%s" % (seg[2], tmp_lun)
        new_path = "/vol/%s/%s" % (seg[2], new_lun)
        try:
            st_tm_mv = self.zapi_client.move_lun(path, tmp_path)
            st_nw_mv = self.zapi_client.move_lun(new_path, path)
            st_del_old = self.zapi_client.destroy_lun(tmp_path)
        except Exception as e:
            if st_tm_mv is None:
                msg = _("Failure staging LUN %s to tmp.")
                raise exception.VolumeBackendAPIException(data=msg % (seg[-1]))
            else:
                if st_nw_mv is None:
                    self.zapi_client.move_lun(tmp_path, path)
                    msg = _("Failure moving new cloned LUN to %s.")
                    raise exception.VolumeBackendAPIException(
                        data=msg % (seg[-1]))
                elif st_del_old is None:
                    LOG.error("Failure deleting staged tmp LUN %s.",
                              tmp_lun)
                else:
                    LOG.error("Unknown exception in"
                              " post clone resize LUN %s.", seg[-1])
                    LOG.error("Exception details: %s", e)

    def _get_lun_block_count(self, path):
        """Gets block counts for the LUN."""
        LOG.debug("Getting LUN block count.")
        lun_infos = self.zapi_client.get_lun_by_args(path=path)
        if not lun_infos:
            seg = path.split('/')
            msg = _('Failure getting LUN info for %s.')
            raise exception.VolumeBackendAPIException(data=msg % seg[-1])
        lun_info = lun_infos[-1]
        bs = int(lun_info.get_child_content('block-size'))
        ls = int(lun_info.get_child_content('size'))
        block_count = ls / bs
        return block_count

    def _check_volume_type_for_lun(self, volume, lun, existing_ref,
                                   extra_specs):
        """Checks if LUN satisfies the volume type."""

    def manage_existing(self, volume, existing_ref):
        """Brings an existing storage object under Cinder management.

        existing_ref can contain source-id or source-name or both.
        source-id: lun uuid.
        source-name: complete lun path eg. /vol/vol0/lun.
        """
        lun = self._get_existing_vol_with_manage_ref(existing_ref)

        extra_specs = na_utils.get_volume_extra_specs(volume)

        self._check_volume_type_for_lun(volume, lun, existing_ref, extra_specs)

        qos_policy_group_info = self._setup_qos_for_volume(volume, extra_specs)
        qos_policy_group_name = (
            na_utils.get_qos_policy_group_name_from_info(
                qos_policy_group_info))
        is_adaptive = na_utils.is_qos_policy_group_spec_adaptive(
            qos_policy_group_info)

        path = lun.get_metadata_property('Path')
        if lun.name == volume['name']:
            new_path = path
            LOG.info("LUN with given ref %s need not be renamed "
                     "during manage operation.", existing_ref)
        else:
            (rest, splitter, name) = path.rpartition('/')
            new_path = '%s/%s' % (rest, volume['name'])
            self.zapi_client.move_lun(path, new_path)
            lun = self._get_existing_vol_with_manage_ref(
                {'source-name': new_path})

        if qos_policy_group_name is not None:
            self.zapi_client.set_lun_qos_policy_group(new_path,
                                                      qos_policy_group_name,
                                                      is_adaptive)
        self._add_lun_to_table(lun)
        LOG.info("Manage operation completed for LUN with new path"
                 " %(path)s and uuid %(uuid)s.",
                 {'path': lun.get_metadata_property('Path'),
                  'uuid': lun.get_metadata_property('UUID')})

        return self._get_volume_model_update(volume)

    def manage_existing_get_size(self, volume, existing_ref):
        """Return size of volume to be managed by manage_existing.

        When calculating the size, round up to the next GB.
        """
        lun = self._get_existing_vol_with_manage_ref(existing_ref)
        return int(math.ceil(float(lun.size) / units.Gi))

    def _get_existing_vol_with_manage_ref(self, existing_ref):
        """Get the corresponding LUN from the storage server."""

        uuid = existing_ref.get('source-id')
        path = existing_ref.get('source-name')

        lun_info = {}
        if path:
            lun_info['path'] = path
        elif uuid:
            if not hasattr(self, 'vserver'):
                reason = _('Volume manage identifier with source-id is only '
                           'supported with clustered Data ONTAP.')
                raise exception.ManageExistingInvalidReference(
                    existing_ref=existing_ref, reason=reason)
            lun_info['uuid'] = uuid
        else:
            reason = _('Volume manage identifier must contain either '
                       'source-id or source-name element.')
            raise exception.ManageExistingInvalidReference(
                existing_ref=existing_ref, reason=reason)

        luns = self.zapi_client.get_lun_by_args(**lun_info)

        for lun in luns:
            netapp_lun = self._extract_lun_info(lun)
            if self._is_lun_valid_on_storage(netapp_lun):
                return netapp_lun

        raise exception.ManageExistingInvalidReference(
            existing_ref=existing_ref,
            reason=(_('LUN not found with given ref %s.') % existing_ref))

    def _is_lun_valid_on_storage(self, lun):
        """Validate lun specific to storage system."""
        return True

    def unmanage(self, volume):
        """Removes the specified volume from Cinder management.

           Does not delete the underlying backend storage object.
        """
        managed_lun = self._get_lun_from_table(volume['name'])
        LOG.info("Unmanaged LUN with current path %(path)s and uuid "
                 "%(uuid)s.",
                 {'path': managed_lun.get_metadata_property('Path'),
                  'uuid': managed_lun.get_metadata_property('UUID')
                  or 'unknown'})

    def initialize_connection_iscsi(self, volume, connector):
        """Driver entry point to attach a volume to an instance.

        Do the LUN masking on the storage system so the initiator can access
        the LUN on the target. Also return the iSCSI properties so the
        initiator can find the LUN. This implementation does not call
        _get_iscsi_properties() to get the properties because cannot store the
        LUN number in the database. We only find out what the LUN number will
        be during this method call so we construct the properties dictionary
        ourselves.
        """

        initiator_name = connector['initiator']
        name = volume['name']
        lun_id = self._map_lun(name, [initiator_name], 'iscsi', None)

        LOG.debug("Mapped LUN %(name)s to the initiator %(initiator_name)s",
                  {'name': name, 'initiator_name': initiator_name})

        target_list = self.zapi_client.get_iscsi_target_details()
        if not target_list:
            raise exception.VolumeBackendAPIException(
                data=_('Failed to get LUN target list for the LUN %s') % name)

        LOG.debug("Successfully fetched target list for LUN %(name)s and "
                  "initiator %(initiator_name)s",
                  {'name': name, 'initiator_name': initiator_name})

        targets = self._get_targets_from_list(target_list)
        if not targets:
            msg = _('Failed to get target portal for the LUN %s')
            raise exception.VolumeBackendAPIException(data=msg % name)
        addresses = [target['address'] for target in targets]
        ports = [target['port'] for target in targets]

        iqn = self.zapi_client.get_iscsi_service_details()
        if not iqn:
            msg = _('Failed to get target IQN for the LUN %s')
            raise exception.VolumeBackendAPIException(data=msg % name)

        properties = na_utils.get_iscsi_connection_properties(lun_id, volume,
                                                              iqn, addresses,
                                                              ports)

        if self.configuration.use_chap_auth:
            chap_username, chap_password = self._configure_chap(initiator_name)
            self._add_chap_properties(properties, chap_username, chap_password)

        return properties

    def _configure_chap(self, initiator_name):
        password = volume_utils.generate_password(na_utils.CHAP_SECRET_LENGTH)
        username = na_utils.DEFAULT_CHAP_USER_NAME

        self.zapi_client.set_iscsi_chap_authentication(initiator_name,
                                                       username,
                                                       password)
        LOG.debug("Set iSCSI CHAP authentication.")

        return username, password

    def _add_chap_properties(self, properties, username, password):
        properties['data']['auth_method'] = 'CHAP'
        properties['data']['auth_username'] = username
        properties['data']['auth_password'] = password
        properties['data']['discovery_auth_method'] = 'CHAP'
        properties['data']['discovery_auth_username'] = username
        properties['data']['discovery_auth_password'] = password

    def _get_targets_from_list(self, target_details_list, filter=None):
        targets = [target for target in target_details_list
                   if (not filter or target['address'] in filter) and
                   target.get('interface-enabled', 'true') == 'true']
        if not targets and len(target_details_list) > 0:
            targets = target_details_list
        return targets

    def _is_multiattached(self, volume, connector):
        """Returns whether the volume is multiattached.

        Returns True if the volume is attached to multiple instances using the
        same initiator as the given one.
        Returns False otherwise.
        """
        if not volume.multiattach or not volume.volume_attachment:
            return False

        same_connector = (True for at in volume.volume_attachment
                          if at.connector and
                          at.connector['initiator'] == connector['initiator'])
        next(same_connector, False)
        return next(same_connector, False)

    def terminate_connection_iscsi(self, volume, connector, **kwargs):
        """Driver entry point to unattach a volume from an instance.

        Unmask the LUN on the storage system so the given initiator can no
        longer access it.
        """
        if connector and self._is_multiattached(volume, connector):
            return

        name = volume['name']
        if connector is None:
            initiators = []
            LOG.debug('Unmapping LUN %(name)s from all initiators',
                      {'name': name})
        else:
            initiators = [connector['initiator']]
            LOG.debug("Unmapping LUN %(name)s from the initiator "
                      "%(initiator_name)s", {'name': name,
                                             'initiator_name': initiators})

        metadata = self._get_lun_attr(name, 'metadata')
        path = metadata['Path']

        self._unmap_lun(path, initiators)

    def initialize_connection_fc(self, volume, connector):
        """Initializes the connection and returns connection info.

        Assign any created volume to a compute node/host so that it can be
        used from that host.

        The driver returns a driver_volume_type of 'fibre_channel'.
        The target_wwn can be a single entry or a list of wwns that
        correspond to the list of remote wwn(s) that will export the volume.
        Example return values:

        .. code-block:: default

            {
                'driver_volume_type': 'fibre_channel',
                'data': {
                    'target_discovered': True,
                    'target_lun': 1,
                    'target_wwn': '500a098280feeba5',
                    'initiator_target_map': {
                        '21000024ff406cc3': ['500a098280feeba5'],
                        '21000024ff406cc2': ['500a098280feeba5']
                    }
                }
            }

        Or

        .. code-block:: default

             {
                'driver_volume_type': 'fibre_channel',
                'data': {
                    'target_discovered': True,
                    'target_lun': 1,
                    'target_wwn': ['500a098280feeba5', '500a098290feeba5',
                                   '500a098190feeba5', '500a098180feeba5'],
                    'initiator_target_map': {
                        '21000024ff406cc3': ['500a098280feeba5',
                                             '500a098290feeba5'],
                        '21000024ff406cc2': ['500a098190feeba5',
                                             '500a098180feeba5']
                    }
                }
            }

        """

        initiators = [fczm_utils.get_formatted_wwn(wwpn)
                      for wwpn in connector['wwpns']]
        volume_name = volume['name']

        lun_id = self._map_lun(volume_name, initiators, 'fcp', None)

        LOG.debug("Mapped LUN %(name)s to the initiator(s) %(initiators)s",
                  {'name': volume_name, 'initiators': initiators})

        target_wwpns, initiator_target_map, num_paths = (
            self._build_initiator_target_map(connector))

        if target_wwpns:
            LOG.debug("Successfully fetched target details for LUN %(name)s "
                      "and initiator(s) %(initiators)s",
                      {'name': volume_name, 'initiators': initiators})
        else:
            raise exception.VolumeBackendAPIException(
                data=_('Failed to get LUN target details for '
                       'the LUN %s') % volume_name)

        target_info = {'driver_volume_type': 'fibre_channel',
                       'data': {'target_discovered': True,
                                'target_lun': int(lun_id),
                                'target_wwn': target_wwpns,
                                'initiator_target_map': initiator_target_map}}

        return target_info

    def terminate_connection_fc(self, volume, connector, **kwargs):
        """Disallow connection from connector.

        Return empty data if other volumes are in the same zone.
        The FibreChannel ZoneManager doesn't remove zones
        if there isn't an initiator_target_map in the
        return of terminate_connection.

        :returns: data - the target_wwns and initiator_target_map if the
                         zone is to be removed, otherwise the same map with
                         an empty dict for the 'data' key
        """

        name = volume['name']
        if connector is None:
            initiators = []
            LOG.debug('Unmapping LUN %(name)s from all initiators',
                      {'name': name})
        else:
            initiators = [fczm_utils.get_formatted_wwn(wwpn)
                          for wwpn in connector['wwpns']]
            LOG.debug("Unmapping LUN %(name)s from the initiators "
                      "%(initiator_name)s", {'name': name,
                                             'initiator_name': initiators})

        metadata = self._get_lun_attr(name, 'metadata')
        path = metadata['Path']

        self._unmap_lun(path, initiators)

        info = {'driver_volume_type': 'fibre_channel',
                'data': {}}

        if connector and not self._has_luns_mapped_to_initiators(initiators):
            # No more exports for this host, so tear down zone.
            LOG.info("Need to remove FC Zone, building initiator target map")

            target_wwpns, initiator_target_map, num_paths = (
                self._build_initiator_target_map(connector))

            info['data'] = {'target_wwn': target_wwpns,
                            'initiator_target_map': initiator_target_map}

        return info

    def _build_initiator_target_map(self, connector):
        """Build the target_wwns and the initiator target map."""

        # get WWPNs from controller and strip colons
        all_target_wwpns = self._get_fc_target_wwpns()
        all_target_wwpns = [six.text_type(wwpn).replace(':', '')
                            for wwpn in all_target_wwpns]

        target_wwpns = []
        init_targ_map = {}
        num_paths = 0

        if self.lookup_service is not None:
            # Use FC SAN lookup to determine which ports are visible.
            dev_map = self.lookup_service.get_device_mapping_from_network(
                connector['wwpns'],
                all_target_wwpns)

            for fabric_name in dev_map:
                fabric = dev_map[fabric_name]
                target_wwpns += fabric['target_port_wwn_list']
                for initiator in fabric['initiator_port_wwn_list']:
                    if initiator not in init_targ_map:
                        init_targ_map[initiator] = []
                    init_targ_map[initiator] += fabric['target_port_wwn_list']
                    init_targ_map[initiator] = list(set(
                        init_targ_map[initiator]))
                    for target in init_targ_map[initiator]:
                        num_paths += 1
            target_wwpns = list(set(target_wwpns))
        else:
            initiator_wwns = connector['wwpns']
            target_wwpns = all_target_wwpns

            for initiator in initiator_wwns:
                init_targ_map[initiator] = target_wwpns

        return target_wwpns, init_targ_map, num_paths

    def _get_backing_flexvol_names(self):
        """Returns a list of backing flexvol names."""
        raise NotImplementedError()



# Copyright (C) 2012 - 2014 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder.api import common


class ViewBuilder(common.ViewBuilder):
    """Model consistencygroup API responses as a python dictionary."""

    _collection_name = "consistencygroups"

    def __init__(self):
        """Initialize view builder."""
        super(ViewBuilder, self).__init__()

    def summary_list(self, request, consistencygroups):
        """Show a list of consistency groups without many details."""
        return self._list_view(self.summary, request, consistencygroups)

    def detail_list(self, request, consistencygroups):
        """Detailed view of a list of consistency groups ."""
        return self._list_view(self.detail, request, consistencygroups)

    def summary(self, request, consistencygroup):
        """Generic, non-detailed view of a consistency group."""
        return {
            'consistencygroup': {
                'id': consistencygroup.id,
                'name': consistencygroup.name
            }
        }

    def detail(self, request, consistencygroup):
        """Detailed view of a single consistency group."""
        try:
            volume_types = (consistencygroup.volume_type_id.split(",")
                            if consistencygroup.volume_type_id else [])
            volume_types = [type_id for type_id in volume_types if type_id]
        except AttributeError:
            try:
                volume_types = [v_type.id for v_type in
                                consistencygroup.volume_types]
            except AttributeError:
                volume_types = []

        return {
            'consistencygroup': {
                'id': consistencygroup.id,
                'status': consistencygroup.status,
                'availability_zone': consistencygroup.availability_zone,
                'created_at': consistencygroup.created_at,
                'name': consistencygroup.name,
                'description': consistencygroup.description,
                'volume_types': volume_types,
            }
        }

    def _list_view(self, func, request, consistencygroups):
        """Provide a view for a list of consistency groups."""
        consistencygroups_list = [
            func(request, consistencygroup)['consistencygroup']
            for consistencygroup in consistencygroups]
        cg_links = self._get_collection_links(request,
                                              consistencygroups,
                                              self._collection_name)
        consistencygroups_dict = dict(consistencygroups=consistencygroups_list)
        if cg_links:
            consistencygroups_dict['consistencygroup_links'] = cg_links

        return consistencygroups_dict



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import os
import shutil
import tempfile
from unittest import mock

from oslo_utils import fileutils
from oslo_utils import timeutils

from cinder.tests.unit import test
from cinder.volume import configuration as conf


class TargetDriverFixture(test.TestCase):
    def setUp(self):
        super(TargetDriverFixture, self).setUp()
        self.configuration = conf.Configuration(None)
        self.configuration.append_config_values = mock.Mock(return_value=0)
        self.configuration.safe_get = mock.Mock(side_effect=self.fake_safe_get)
        self.configuration.target_ip_address = '10.9.8.7'
        self.configuration.target_port = 3260

        self.fake_volumes_dir = tempfile.mkdtemp()
        fileutils.ensure_tree(self.fake_volumes_dir)

        self.fake_project_id = 'ed2c1fd4-5fc0-11e4-aa15-123b93f75cba'
        self.fake_project_id_2 = 'ed2c1fd4-5fc0-11e4-aa15-123b93f75cba'
        self.fake_volume_id = 'ed2c2222-5fc0-11e4-aa15-123b93f75cba'

        self.addCleanup(self._cleanup)

        self.testvol =\
            {'project_id': self.fake_project_id,
             'name': 'testvol',
             'size': 1,
             'id': self.fake_volume_id,
             'volume_type_id': None,
             'provider_location': '10.10.7.1:3260 '
                                  'iqn.2010-10.org.openstack:'
                                  'volume-%s 0' % self.fake_volume_id,
             'provider_auth': 'CHAP stack-1-a60e2611875f40199931f2'
                              'c76370d66b 2FE0CQ8J196R',
             'provider_geometry': '512 512',
             'created_at': timeutils.utcnow(),
             'host': 'fake_host@lvm#lvm'}

        self.testvol_no_prov_loc = copy.copy(self.testvol)
        self.testvol_no_prov_loc['provider_location'] = None

        self.iscsi_target_prefix = 'iqn.2010-10.org.openstack:'
        self.target_string = ('127.0.0.1:3260,1 ' +
                              self.iscsi_target_prefix +
                              'volume-%s' % self.testvol['id'])

        self.testvol_2 =\
            {'project_id': self.fake_project_id_2,
             'name': 'testvol2',
             'size': 1,
             'id': self.fake_volume_id,
             'volume_type_id': None,
             'provider_location': ('%(ip)s:%(port)d%(iqn)svolume-%(vol)s 2' %
                                   {'ip': self.configuration.target_ip_address,
                                    'port': self.configuration.target_port,
                                    'iqn': self.iscsi_target_prefix,
                                    'vol': self.fake_volume_id}),
             'provider_auth': 'CHAP stack-1-a60e2611875f40199931f2'
                              'c76370d66b 2FE0CQ8J196R',
             'provider_geometry': '512 512',
             'created_at': timeutils.utcnow(),
             'host': 'fake_host@lvm#lvm'}

        self.expected_iscsi_properties = \
            {'auth_method': 'CHAP',
             'auth_password': '2FE0CQ8J196R',
             'auth_username': 'stack-1-a60e2611875f40199931f2c76370d66b',
             'encrypted': False,
             'logical_block_size': '512',
             'physical_block_size': '512',
             'target_discovered': False,
             'target_iqn': 'iqn.2010-10.org.openstack:volume-%s' %
                           self.fake_volume_id,
             'target_lun': 0,
             'target_portal': '10.10.7.1:3260',
             'volume_id': self.fake_volume_id}

        self.VOLUME_ID = '83c2e877-feed-46be-8435-77884fe55b45'
        self.VOLUME_NAME = 'volume-' + self.VOLUME_ID
        self.test_vol = (self.iscsi_target_prefix +
                         self.VOLUME_NAME)

    def _cleanup(self):
        if os.path.exists(self.fake_volumes_dir):
            shutil.rmtree(self.fake_volumes_dir)

    def fake_safe_get(self, value):
        if value == 'volumes_dir':
            return self.fake_volumes_dir
        elif value == 'target_protocol':
            return self.configuration.target_protocol
        elif value == 'target_prefix':
            return self.iscsi_target_prefix



# Copyright 2015 Chelsio Communications Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import sys
from unittest import mock

from cinder import context
from cinder.tests.unit.targets import targets_fixture as tf
from cinder.tests.unit import test
from cinder import utils
from cinder.volume.targets import cxt


class TestCxtAdmDriver(tf.TargetDriverFixture):
    def setUp(self):
        super(TestCxtAdmDriver, self).setUp()
        self.cxt_subdir = cxt.CxtAdm.cxt_subdir
        self.target = cxt.CxtAdm(root_helper=utils.get_root_helper(),
                                 configuration=self.configuration)
        self.VG = 'stack-volumes-lvmdriver-1'
        self.fake_iscsi_scan = \
            ('\n'
             'TARGET: iqn.2010-10.org.openstack:%(vol)s, id=1, login_ip=0\n'
             '        PortalGroup=1@10.9.8.7:3260,timeout=0\n'
             '        TargetDevice=/dev/%(vg)s/%(vol)s'
             ',BLK,PROD=CHISCSI '
             'Target,SN=0N0743000000000,ID=0D074300000000000000000,'
             'WWN=:W00743000000000\n'
             % {'vol': self.VOLUME_NAME, 'vg': self.VG})

    def test_get_target(self):
        with mock.patch.object(self.target, '_get_volumes_dir',
                               return_value=self.fake_volumes_dir),\
            mock.patch('cinder.utils.execute',
                       return_value=(self.fake_iscsi_scan, None)) as m_exec:
            self.assertEqual(
                '1',
                self.target._get_target(
                    'iqn.2010-10.org.openstack:volume-%s' % self.VOLUME_ID
                )
            )
            self.assertTrue(m_exec.called)

    @test.testtools.skipIf(sys.platform == "darwin", "SKIP on OSX")
    @mock.patch('cinder.volume.targets.cxt.CxtAdm._get_target',
                return_value=1)
    @mock.patch('cinder.utils.execute')
    def test_create_iscsi_target(self, mock_execute, mock_get_targ):
        mock_execute.return_value = ('', '')
        with mock.patch.object(self.target, '_get_volumes_dir') as mock_get:
            mock_get.return_value = self.fake_volumes_dir
            self.assertEqual(
                1,
                self.target.create_iscsi_target(
                    self.test_vol,
                    1,
                    0,
                    self.fake_volumes_dir,
                    portals_ips=[self.configuration.target_ip_address]))
            self.assertTrue(mock_get.called)
            self.assertTrue(mock_execute.called)
            self.assertTrue(mock_get_targ.called)

    @test.testtools.skipIf(sys.platform == "darwin", "SKIP on OSX")
    @mock.patch('cinder.volume.targets.cxt.CxtAdm._get_target',
                return_value=1)
    @mock.patch('cinder.utils.execute', return_value=('fake out', 'fake err'))
    def test_create_iscsi_target_port_ips(self, mock_execute, mock_get_targ):
        ips = ['10.0.0.15', '127.0.0.1']
        port = 3261
        mock_execute.return_value = ('', '')
        with mock.patch.object(self.target, '_get_volumes_dir') as mock_get:
            mock_get.return_value = self.fake_volumes_dir
            test_vol = 'iqn.2010-10.org.openstack:'\
                       'volume-83c2e877-feed-46be-8435-77884fe55b45'
            self.assertEqual(
                1,
                self.target.create_iscsi_target(
                    test_vol,
                    1,
                    0,
                    self.fake_volumes_dir,
                    portals_port=port,
                    portals_ips=ips))

            self.assertTrue(mock_get.called)
            self.assertTrue(mock_execute.called)
            self.assertTrue(mock_get_targ.called)

            file_path = os.path.join(self.fake_volumes_dir,
                                     test_vol.split(':')[1])

            expected_cfg = {
                'name': test_vol,
                'device': self.fake_volumes_dir,
                'ips': ','.join(map(lambda ip: '%s:%s' % (ip, port), ips)),
                'spaces': ' ' * 14,
                'spaces2': ' ' * 23}

            expected_file = ('\n%(spaces)starget:'
                             '\n%(spaces2)sTargetName=%(name)s'
                             '\n%(spaces2)sTargetDevice=%(device)s'
                             '\n%(spaces2)sPortalGroup=1@%(ips)s'
                             '\n%(spaces)s   ') % expected_cfg

            with open(file_path, 'r') as cfg_file:
                result = cfg_file.read()
                self.assertEqual(expected_file, result)

    @test.testtools.skipIf(sys.platform == "darwin", "SKIP on OSX")
    @mock.patch('cinder.volume.targets.cxt.CxtAdm._get_target',
                return_value=1)
    @mock.patch('cinder.utils.execute', return_value=('fake out', 'fake err'))
    def test_create_iscsi_target_already_exists(self, mock_execute,
                                                mock_get_targ):
        with mock.patch.object(self.target, '_get_volumes_dir') as mock_get:
            mock_get.return_value = self.fake_volumes_dir
            self.assertEqual(
                1,
                self.target.create_iscsi_target(
                    self.test_vol,
                    1,
                    0,
                    self.fake_volumes_dir,
                    portals_ips=[self.configuration.target_ip_address]))
            self.assertTrue(mock_get.called)
            self.assertTrue(mock_get_targ.called)
            self.assertTrue(mock_execute.called)

    @test.testtools.skipIf(sys.platform == "darwin", "SKIP on OSX")
    @mock.patch('cinder.volume.targets.cxt.CxtAdm._get_target',
                return_value=1)
    @mock.patch('cinder.utils.execute')
    @mock.patch.object(cxt.CxtAdm, '_get_target_chap_auth')
    def test_create_export(self, mock_chap, mock_execute,
                           mock_get_targ):
        mock_execute.return_value = ('', '')
        mock_chap.return_value = ('QZJbisGmn9AL954FNF4D',
                                  'P68eE7u9eFqDGexd28DQ')
        with mock.patch.object(self.target, '_get_volumes_dir') as mock_get:
            mock_get.return_value = self.fake_volumes_dir

            expected_result = {'location': '10.9.8.7:3260,1 '
                               'iqn.2010-10.org.openstack:testvol 0',
                               'auth': 'CHAP '
                               'QZJbisGmn9AL954FNF4D P68eE7u9eFqDGexd28DQ'}

            ctxt = context.get_admin_context()
            self.assertEqual(expected_result,
                             self.target.create_export(ctxt,
                                                       self.testvol,
                                                       self.fake_volumes_dir))
            self.assertTrue(mock_get.called)
            self.assertTrue(mock_execute.called)

    @mock.patch('cinder.volume.targets.cxt.CxtAdm._get_target_chap_auth')
    def test_ensure_export(self, mock_get_chap):
        fake_creds = ('asdf', 'qwert')
        mock_get_chap.return_value = fake_creds
        ctxt = context.get_admin_context()
        with mock.patch.object(self.target, 'create_iscsi_target'):
            self.target.ensure_export(ctxt,
                                      self.testvol,
                                      self.fake_volumes_dir)
            self.target.create_iscsi_target.assert_called_once_with(
                'iqn.2010-10.org.openstack:testvol',
                1, 0, self.fake_volumes_dir, fake_creds,
                check_exit_code=False,
                old_name=None,
                portals_ips=[self.configuration.target_ip_address],
                portals_port=self.configuration.target_port)



# Stubs for oslo_i18n._gettextutils (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

def install(domain: Any) -> None: ...
def get_available_languages(domain: Any): ...



# Copyright 2012, Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Client side of the scheduler manager RPC API.
"""

from oslo_serialization import jsonutils
from oslo_utils import timeutils

from cinder.common import constants
from cinder import rpc


class SchedulerAPI(rpc.RPCAPI):
    """Client side of the scheduler RPC API.

    API version history:

    .. code-block:: none

        1.0 - Initial version.
        1.1 - Add create_volume() method
        1.2 - Add request_spec, filter_properties arguments to
              create_volume()
        1.3 - Add migrate_volume_to_host() method
        1.4 - Add retype method
        1.5 - Add manage_existing method
        1.6 - Add create_consistencygroup method
        1.7 - Add get_active_pools method
        1.8 - Add sending object over RPC in create_consistencygroup method
        1.9 - Adds support for sending objects over RPC in create_volume()
        1.10 - Adds support for sending objects over RPC in retype()
        1.11 - Adds support for sending objects over RPC in
               migrate_volume_to_host()

        ... Mitaka supports messaging 1.11. Any changes to existing methods in
        1.x after this point should be done so that they can handle version cap
        set to 1.11.

        2.0 - Remove 1.x compatibility
        2.1 - Adds support for sending objects over RPC in manage_existing()
        2.2 - Sends request_spec as object in create_volume()
        2.3 - Add create_group method

        ... Newton supports messaging 2.3. Any changes to existing methods in
        2.x after this point should be done so that they can handle version cap
        set to 2.3.

        3.0 - Remove 2.x compatibility
        3.1 - Adds notify_service_capabilities()
        3.2 - Adds extend_volume()
        3.3 - Add cluster support to migrate_volume, and to
              update_service_capabilities and send the timestamp from the
              capabilities.
        3.4 - Adds work_cleanup and do_cleanup methods.
        3.5 - Make notify_service_capabilities support A/A
        3.6 - Removed create_consistencygroup method
        3.7 - Adds set_log_levels and get_log_levels
        3.8 - Addds ``valid_host_capacity`` method
        3.9 - Adds create_snapshot method
        3.10 - Adds backup_id to create_volume method.
        3.11 - Adds manage_existing_snapshot method.
        3.12 - Adds create_backup method.
    """

    RPC_API_VERSION = '3.12'
    RPC_DEFAULT_VERSION = '3.0'
    TOPIC = constants.SCHEDULER_TOPIC
    BINARY = 'cinder-scheduler'

    def create_group(self, ctxt, group, group_spec=None,
                     request_spec_list=None, group_filter_properties=None,
                     filter_properties_list=None):
        cctxt = self._get_cctxt()
        request_spec_p_list = [jsonutils.to_primitive(rs)
                               for rs in request_spec_list]
        group_spec_p = jsonutils.to_primitive(group_spec)
        msg_args = {
            'group': group, 'group_spec': group_spec_p,
            'request_spec_list': request_spec_p_list,
            'group_filter_properties': group_filter_properties,
            'filter_properties_list': filter_properties_list,
        }

        cctxt.cast(ctxt, 'create_group', **msg_args)

    def create_volume(self, ctxt, volume, snapshot_id=None, image_id=None,
                      request_spec=None, filter_properties=None,
                      backup_id=None):
        volume.create_worker()
        cctxt = self._get_cctxt()
        msg_args = {'snapshot_id': snapshot_id, 'image_id': image_id,
                    'request_spec': request_spec,
                    'filter_properties': filter_properties,
                    'volume': volume, 'backup_id': backup_id}
        if not self.client.can_send_version('3.10'):
            msg_args.pop('backup_id')
        return cctxt.cast(ctxt, 'create_volume', **msg_args)

    @rpc.assert_min_rpc_version('3.8')
    def validate_host_capacity(self, ctxt, backend, request_spec,
                               filter_properties=None):
        msg_args = {'request_spec': request_spec,
                    'filter_properties': filter_properties, 'backend': backend}
        cctxt = self._get_cctxt()
        return cctxt.call(ctxt, 'validate_host_capacity', **msg_args)

    @rpc.assert_min_rpc_version('3.9')
    def create_snapshot(self, ctxt, volume, snapshot, backend,
                        request_spec=None, filter_properties=None):
        cctxt = self._get_cctxt()
        msg_args = {'request_spec': request_spec,
                    'filter_properties': filter_properties,
                    'volume': volume,
                    'snapshot': snapshot,
                    'backend': backend}
        return cctxt.cast(ctxt, 'create_snapshot', **msg_args)

    def migrate_volume(self, ctxt, volume, backend, force_copy=False,
                       request_spec=None, filter_properties=None):
        request_spec_p = jsonutils.to_primitive(request_spec)
        msg_args = {'request_spec': request_spec_p,
                    'filter_properties': filter_properties, 'volume': volume}
        version = '3.3'
        if self.client.can_send_version(version):
            msg_args['backend'] = backend
            msg_args['force_copy'] = force_copy
            method = 'migrate_volume'
        else:
            version = '3.0'
            msg_args['host'] = backend
            msg_args['force_host_copy'] = force_copy
            method = 'migrate_volume_to_host'

        cctxt = self._get_cctxt(version=version)
        return cctxt.cast(ctxt, method, **msg_args)

    def retype(self, ctxt, volume, request_spec=None, filter_properties=None):
        cctxt = self._get_cctxt()
        request_spec_p = jsonutils.to_primitive(request_spec)
        msg_args = {'request_spec': request_spec_p,
                    'filter_properties': filter_properties, 'volume': volume}
        return cctxt.cast(ctxt, 'retype', **msg_args)

    def manage_existing(self, ctxt, volume, request_spec=None,
                        filter_properties=None):
        cctxt = self._get_cctxt()
        request_spec_p = jsonutils.to_primitive(request_spec)
        msg_args = {
            'request_spec': request_spec_p,
            'filter_properties': filter_properties, 'volume': volume,
        }
        return cctxt.cast(ctxt, 'manage_existing', **msg_args)

    @rpc.assert_min_rpc_version('3.11')
    def manage_existing_snapshot(self, ctxt, volume, snapshot, ref,
                                 request_spec=None, filter_properties=None):
        cctxt = self._get_cctxt()
        request_spec_p = jsonutils.to_primitive(request_spec)
        msg_args = {
            'request_spec': request_spec_p,
            'filter_properties': filter_properties,
            'volume': volume,
            'snapshot': snapshot,
            'ref': ref,
        }
        return cctxt.cast(ctxt, 'manage_existing_snapshot', **msg_args)

    @rpc.assert_min_rpc_version('3.2')
    def extend_volume(self, ctxt, volume, new_size, reservations,
                      request_spec, filter_properties=None):
        cctxt = self._get_cctxt()

        request_spec_p = jsonutils.to_primitive(request_spec)
        msg_args = {
            'volume': volume,
            'new_size': new_size,
            'reservations': reservations,
            'request_spec': request_spec_p,
            'filter_properties': filter_properties,
        }

        return cctxt.cast(ctxt, 'extend_volume', **msg_args)

    def get_pools(self, ctxt, filters=None):
        cctxt = self._get_cctxt()
        return cctxt.call(ctxt, 'get_pools', filters=filters)

    @staticmethod
    def prepare_timestamp(timestamp):
        timestamp = timestamp or timeutils.utcnow()
        return jsonutils.to_primitive(timestamp)

    def update_service_capabilities(self, ctxt, service_name, host,
                                    capabilities, cluster_name,
                                    timestamp=None):
        msg_args = dict(service_name=service_name, host=host,
                        capabilities=capabilities)

        version = '3.3'
        # If server accepts timestamping the capabilities and the cluster name
        if self.client.can_send_version(version):
            # Serialize the timestamp
            msg_args.update(cluster_name=cluster_name,
                            timestamp=self.prepare_timestamp(timestamp))
        else:
            version = '3.0'

        cctxt = self._get_cctxt(fanout=True, version=version)
        cctxt.cast(ctxt, 'update_service_capabilities', **msg_args)

    @rpc.assert_min_rpc_version('3.1')
    def notify_service_capabilities(self, ctxt, service_name,
                                    backend, capabilities, timestamp=None):
        parameters = {'service_name': service_name,
                      'capabilities': capabilities}
        if self.client.can_send_version('3.5'):
            version = '3.5'
            parameters.update(backend=backend,
                              timestamp=self.prepare_timestamp(timestamp))
        else:
            version = '3.1'
            parameters['host'] = backend

        cctxt = self._get_cctxt(version=version)
        cctxt.cast(ctxt, 'notify_service_capabilities', **parameters)

    @rpc.assert_min_rpc_version('3.4')
    def work_cleanup(self, ctxt, cleanup_request):
        """Generate individual service cleanup requests from user request."""
        cctxt = self.client.prepare(version='3.4')
        # Response will have services that are receiving the cleanup request
        # and services that couldn't receive it since they are down.
        return cctxt.call(ctxt, 'work_cleanup',
                          cleanup_request=cleanup_request)

    @rpc.assert_min_rpc_version('3.4')
    def do_cleanup(self, ctxt, cleanup_request):
        """Perform this scheduler's resource cleanup as per cleanup_request."""
        cctxt = self.client.prepare(version='3.4')
        cctxt.cast(ctxt, 'do_cleanup', cleanup_request=cleanup_request)

    @rpc.assert_min_rpc_version('3.7')
    def set_log_levels(self, context, service, log_request):
        cctxt = self._get_cctxt(server=service.host, version='3.7')
        cctxt.cast(context, 'set_log_levels', log_request=log_request)

    @rpc.assert_min_rpc_version('3.7')
    def get_log_levels(self, context, service, log_request):
        cctxt = self._get_cctxt(server=service.host, version='3.7')
        return cctxt.call(context, 'get_log_levels', log_request=log_request)

    @rpc.assert_min_rpc_version('3.12')
    def create_backup(self, ctxt, backup):
        cctxt = self._get_cctxt()
        msg_args = {'backup': backup}
        return cctxt.cast(ctxt, 'create_backup', **msg_args)



# Copyright 2021 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sqlalchemy as sa


def upgrade(migrate_engine):
    """Update volumes and snapshots tables with use_quota field.

    Add use_quota field to both volumes and snapshots table to fast and easily
    identify resources that must be counted for quota usages.
    """
    # Existing resources will be left with None value to allow rolling upgrades
    # with the online data migration pattern, since they will identify the
    # resources that don't have the field set/known yet.
    meta = sa.MetaData(bind=migrate_engine)
    for table_name in ('volumes', 'snapshots'):
        table = sa.Table(table_name, meta, autoload=True)

        if not hasattr(table.c, 'use_quota'):
            column = sa.Column('use_quota', sa.Boolean, nullable=True)
            table.create_column(column)



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Database setup and migration commands."""

import os

from alembic import command as alembic_api
from alembic import config as alembic_config
from alembic import migration as alembic_migration
from migrate import exceptions as migrate_exceptions
from migrate.versioning import api as migrate_api
from migrate.versioning import repository as migrate_repo
from oslo_config import cfg
from oslo_db import options
from oslo_log import log as logging

from cinder.db.sqlalchemy import api as db_api

options.set_defaults(cfg.CONF)

LOG = logging.getLogger(__name__)

MIGRATE_INIT_VERSION = 134
MIGRATE_MIGRATIONS_PATH = ALEMBIC_INIT_VERSION = '921e1a36b076'


def _find_migrate_repo():
    """Get the project's change script repository

    :returns: An instance of ``migrate.versioning.repository.Repository``
    """
    path = os.path.join(
        os.path.abspath(os.path.dirname(__file__)), 'legacy_migrations',
    )
    return migrate_repo.Repository(path)


def _find_alembic_conf():
    """Get the project's alembic configuration

    :returns: An instance of ``alembic.config.Config``
    """
    path = os.path.join(
        os.path.abspath(os.path.dirname(__file__)), 'alembic.ini')

    config = alembic_config.Config(os.path.abspath(path))
    # we don't want to use the logger configuration from the file, which is
    # only really intended for the CLI
    # https://stackoverflow.com/a/42691781/613428
    config.attributes['configure_logger'] = False

    return config


def _is_database_under_migrate_control(engine, repository):
    try:
        migrate_api.db_version(engine, repository)
        return True
    except migrate_exceptions.DatabaseNotControlledError:
        return False


def _is_database_under_alembic_control(engine):
    with engine.connect() as conn:
        context = alembic_migration.MigrationContext.configure(conn)
        return bool(context.get_current_revision())


def _init_alembic_on_legacy_database(engine, repository, config):
    """Init alembic in an existing environment with sqlalchemy-migrate."""
    LOG.info(
        'The database is still under sqlalchemy-migrate control; '
        'applying any remaining sqlalchemy-migrate-based migrations '
        'and fake applying the initial alembic migration'
    )
    migrate_api.upgrade(engine, repository)

    # re-use the connection rather than creating a new one
    with engine.begin() as connection:
        config.attributes['connection'] = connection
        alembic_api.stamp(config, ALEMBIC_INIT_VERSION)


def _upgrade_alembic(engine, config, version):
    # re-use the connection rather than creating a new one
    with engine.begin() as connection:
        config.attributes['connection'] = connection
        alembic_api.upgrade(config, version or 'head')


def db_version():
    """Get database version."""
    repository = _find_migrate_repo()
    engine = db_api.get_engine()

    migrate_version = None
    if _is_database_under_migrate_control(engine, repository):
        migrate_version = migrate_api.db_version(engine, repository)

    alembic_version = None
    if _is_database_under_alembic_control(engine):
        with engine.connect() as conn:
            m_context = alembic_migration.MigrationContext.configure(conn)
            alembic_version = m_context.get_current_revision()

    return alembic_version or migrate_version


def db_sync(version=None, engine=None):
    """Migrate the database to `version` or the most recent version.

    We're currently straddling two migration systems, sqlalchemy-migrate and
    alembic. This handles both by ensuring we switch from one to the other at
    the appropriate moment.
    """

    # if the user requested a specific version, check if it's an integer: if
    # so, we're almost certainly in sqlalchemy-migrate land and won't support
    # that
    if version is not None and version.isdigit():
        raise ValueError(
            'You requested an sqlalchemy-migrate database version; this is '
            'no longer supported'
        )

    if engine is None:
        engine = db_api.get_engine()

    repository = _find_migrate_repo()
    config = _find_alembic_conf()

    # discard the URL encoded in alembic.ini in favour of the URL configured
    # for the engine by the database fixtures, casting from
    # 'sqlalchemy.engine.url.URL' to str in the process. This returns a
    # RFC-1738 quoted URL, which means that a password like "foo@" will be
    # turned into "foo%40". This in turns causes a problem for
    # set_main_option() because that uses ConfigParser.set, which (by design)
    # uses *python* interpolation to write the string out ... where "%" is the
    # special python interpolation character! Avoid this mismatch by quoting
    # all %'s for the set below.
    engine_url = str(engine.url).replace('%', '%%')
    config.set_main_option('sqlalchemy.url', str(engine_url))

    # if we're in a deployment where sqlalchemy-migrate is already present,
    # then apply all the updates for that and fake apply the initial alembic
    # migration; if we're not then 'upgrade' will take care of everything
    # this should be a one-time operation
    if (
        _is_database_under_migrate_control(engine, repository) and
        not _is_database_under_alembic_control(engine)
    ):
        _init_alembic_on_legacy_database(engine, repository, config)

    # apply anything later
    LOG.info('Applying migration(s)')
    _upgrade_alembic(engine, config, version)
    LOG.info('Migration(s) applied')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_utils import timeutils

from cinder import context
from cinder.tests.unit.targets import targets_fixture as tf
from cinder import utils
from cinder.volume.targets import nvmet


class TestNVMETDriver(tf.TargetDriverFixture):

    def setUp(self):
        super(TestNVMETDriver, self).setUp()

        self.configuration.target_protocol = 'nvmet_rdma'
        self.target = nvmet.NVMET(root_helper=utils.get_root_helper(),
                                  configuration=self.configuration)

        self.target_ip = '192.168.0.1'
        self.target_port = '1234'
        self.nvmet_subsystem_name = self.configuration.target_prefix
        self.nvmet_ns_id = self.configuration.nvmet_ns_id
        self.nvmet_port_id = self.configuration.nvmet_port_id
        self.nvme_transport_type = 'rdma'

        self.fake_volume_id = 'c446b9a2-c968-4260-b95f-a18a7b41c004'
        self.testvol_path = (
            '/dev/stack-volumes-lvmdriver-1/volume-%s' % self.fake_volume_id)
        self.fake_project_id = 'ed2c1fd4-5555-1111-aa15-123b93f75cba'
        self.testvol = (
            {'project_id': self.fake_project_id,
             'name': 'testvol',
             'size': 1,
             'id': self.fake_volume_id,
             'volume_type_id': None,
             'provider_location': self.target.get_nvmeof_location(
                 "nqn.%s-%s" % (self.nvmet_subsystem_name,
                                self.fake_volume_id),
                 self.target_ip, self.target_port, self.nvme_transport_type,
                 self.nvmet_ns_id),
             'provider_auth': None,
             'provider_geometry': None,
             'created_at': timeutils.utcnow(),
             'host': 'fake_host@lvm#lvm'})

    @mock.patch.object(nvmet.NVMET, '_get_nvmf_subsystem')
    @mock.patch.object(nvmet.NVMET, '_get_available_nvmf_subsystems')
    @mock.patch.object(nvmet.NVMET, '_add_nvmf_subsystem')
    def test_create_export(self, mock_add_nvmf_subsystem,
                           mock_get_available_nvmf_subsystems,
                           mock_get_nvmf_subsystem):

        mock_testvol = self.testvol
        mock_testvol_path = self.testvol_path
        ctxt = context.get_admin_context()
        mock_get_available_nvmf_subsystems.return_value = {
            "subsystems": [],
            "hosts": [],
            "ports": [
                {"subsystems": [],
                 "referrals": [],
                 "portid": 1,
                 "addr":
                 {"treq": "not specified",
                          "trtype": "rdma",
                          "adrfam": "ipv4",
                          "trsvcid": self.target_port,
                          "traddr":
                              self.target_ip
                  }
                 }]
        }
        mock_get_nvmf_subsystem.return_value = (
            "nqn.%s-%s" % (self.nvmet_subsystem_name,
                           mock_testvol['id']))

        mock_add_nvmf_subsystem.return_value = (
            "nqn.%s-%s" % (self.nvmet_subsystem_name,
                           mock_testvol['id']))

        expected_return = {
            'location': self.target.get_nvmeof_location(
                mock_add_nvmf_subsystem.return_value, self.target_ip,
                self.target_port, self.nvme_transport_type, self.nvmet_ns_id),
            'auth': ''
        }

        self.target.target_ip = self.target_ip
        self.target.target_port = self.target_port
        self.assertEqual(expected_return,
                         self.target.create_export(
                             ctxt, mock_testvol,
                             mock_testvol_path))

    @mock.patch.object(nvmet.NVMET, '_get_nvmf_subsystem')
    @mock.patch.object(nvmet.NVMET, '_get_available_nvmf_subsystems')
    @mock.patch.object(nvmet.NVMET, '_add_nvmf_subsystem')
    def test_create_export_with_error_add_nvmf_subsystem(
            self,
            mock_add_nvmf_subsystem,
            mock_get_available_nvmf_subsystems,
            mock_get_nvmf_subsystem):

        mock_testvol = self.testvol
        mock_testvol_path = self.testvol_path
        ctxt = context.get_admin_context()
        mock_get_available_nvmf_subsystems.return_value = {
            "subsystems": [],
            "hosts": [],
            "ports": [
                {"subsystems": [],
                 "referrals": [],
                 "portid": 1,
                 "addr":
                 {"treq": "not specified",
                          "trtype": "rdma",
                          "adrfam": "ipv4",
                          "trsvcid": self.target_port,
                          "traddr":
                              self.target_ip
                  }
                 }]
        }
        mock_get_nvmf_subsystem.return_value = None

        mock_add_nvmf_subsystem.return_value = None

        self.target.target_ip = self.target_ip
        self.target.target_port = self.target_port
        self.assertRaises(nvmet.NVMETTargetAddError,
                          self.target.create_export,
                          ctxt,
                          mock_testvol,
                          mock_testvol_path)

    @mock.patch.object(nvmet.NVMET, '_get_nvmf_subsystem')
    @mock.patch.object(nvmet.NVMET, '_get_available_nvmf_subsystems')
    @mock.patch.object(nvmet.NVMET, '_delete_nvmf_subsystem')
    def test_remove_export(self, mock_delete_nvmf_subsystem,
                           mock_get_available_nvmf_subsystems,
                           mock_get_nvmf_subsystem):
        mock_testvol = self.testvol
        mock_testvol_path = self.testvol_path
        ctxt = context.get_admin_context()
        mock_get_available_nvmf_subsystems.return_value = {
            "subsystems": [
                {"allowed_hosts": [],
                 "nqn": "nqn.%s-%s" % (
                     self.nvmet_subsystem_name,
                     mock_testvol['id']),
                 "attr": {"allow_any_host": "1"},
                 "namespaces": [
                 {"device":
                  {"path": mock_testvol_path,
                   "nguid":
                   "86fab0e0-825d-4f25-a449-28b93c5e8dd6"
                   },
                  "enable": 1, "nsid":
                  self.nvmet_ns_id,
                  }]}],
            "hosts": [],
            "ports": [
                {"subsystems": [
                    "nqn.%s-%s" % (self.nvmet_subsystem_name,
                                   mock_testvol['id'])],
                 "referrals": [],
                 "portid": self.nvmet_port_id,
                 "addr":
                 {"treq": "not specified",
                  "trtype": "rdma",
                  "adrfam": "ipv4",
                  "trsvcid": self.target_port,
                  "traddr": self.target_ip}}
            ]
        }

        mock_get_nvmf_subsystem.return_value = (
            "nqn.%s-%s" % (self.nvmet_subsystem_name,
                           mock_testvol['id']))
        mock_delete_nvmf_subsystem.return_value = (
            "nqn.%s-%s" % (self.nvmet_subsystem_name,
                           mock_testvol['id']))
        expected_return = mock_delete_nvmf_subsystem.return_value
        self.assertEqual(expected_return,
                         self.target.remove_export(ctxt, mock_testvol))

    @mock.patch.object(nvmet.NVMET, '_get_nvmf_subsystem')
    @mock.patch.object(nvmet.NVMET, '_get_available_nvmf_subsystems')
    def test_remove_export_with_empty_subsystems(
            self,
            mock_get_available_nvmf_subsystems,
            mock_get_nvmf_subsystem):
        mock_testvol = self.testvol
        ctxt = context.get_admin_context()
        mock_get_available_nvmf_subsystems.return_value = {
            "subsystems": [],
            "hosts": [],
            "ports": []
        }
        mock_get_nvmf_subsystem.return_value = None
        self.assertIsNone(self.target.remove_export(ctxt, mock_testvol))

    @mock.patch.object(nvmet.NVMET, '_get_nvmf_subsystem')
    @mock.patch.object(nvmet.NVMET, '_get_available_nvmf_subsystems')
    @mock.patch.object(nvmet.NVMET, '_delete_nvmf_subsystem')
    def test_remove_export_with_delete_nvmf_subsystem_fails(
            self,
            moc_delete_nvmf_subsystem,
            mock_get_available_nvmf_subsystems,
            mock_get_nvmf_subsystem):
        mock_testvol = self.testvol
        mock_testvol_path = self.testvol_path
        ctxt = context.get_admin_context()
        mock_get_available_nvmf_subsystems.return_value = {
            "subsystems": [
                {"allowed_hosts": [],
                 "nqn": "nqn.%s-%s" % (
                     self.nvmet_subsystem_name,
                     mock_testvol['id']),
                 "attr": {"allow_any_host": "1"},
                 "namespaces": [
                 {"device":
                  {"path": mock_testvol_path,
                   "nguid":
                   "86fab0e0-825d-4f25-a449-28b93c5e8dd6"
                   },
                  "enable": 1, "nsid":
                  self.nvmet_ns_id,
                  }]}],
            "hosts": [],
            "ports": [
                {"subsystems": [
                    "nqn.%s-%s" % (self.nvmet_subsystem_name,
                                   mock_testvol['id'])],
                 "referrals": [],
                 "portid": self.nvmet_port_id,
                 "addr":
                 {"treq": "not specified",
                  "trtype": "rdma",
                  "adrfam": "ipv4",
                  "trsvcid": self.target_port,
                  "traddr": self.target_ip}}
            ]
        }
        mock_get_nvmf_subsystem.return_value = (
            "nqn.%s-%s" % (self.nvmet_subsystem_name,
                           mock_testvol['id']))
        moc_delete_nvmf_subsystem.return_value = None
        self.assertRaises(nvmet.NVMETTargetDeleteError,
                          self.target.remove_export,
                          ctxt,
                          mock_testvol)



# Copyright (c) 2013 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The volume encryption metadata extension."""

from cinder.api import extensions
from cinder.api.openstack import wsgi
from cinder import db
from cinder import objects
from cinder.policies import volumes as policy


class VolumeEncryptionMetadataController(wsgi.Controller):
    """The volume encryption metadata API extension."""

    def index(self, req, volume_id):
        """Returns the encryption metadata for a given volume."""
        context = req.environ['cinder.context']
        volume = objects.Volume.get_by_id(context, volume_id)
        context.authorize(policy.ENCRYPTION_METADATA_POLICY,
                          target_obj=volume)
        return db.volume_encryption_metadata_get(context, volume_id)

    def show(self, req, volume_id, id):
        """Return a single encryption item."""
        encryption_item = self.index(req, volume_id)
        if encryption_item is not None:
            return encryption_item[id]
        else:
            return None


class Volume_encryption_metadata(extensions.ExtensionDescriptor):
    """Volume encryption metadata retrieval support."""

    name = "VolumeEncryptionMetadata"
    alias = "os-volume-encryption-metadata"
    updated = "2013-07-10T00:00:00+00:00"

    def get_resources(self):
        resources = []
        res = extensions.ResourceExtension(
            'encryption', VolumeEncryptionMetadataController(),
            parent=dict(member_name='volume', collection_name='volumes'))
        resources.append(res)
        return resources



#    (c) Copyright 2014 Brocade Communications Systems Inc.
#    All Rights Reserved.
#
#    Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""
Base Zone Driver is responsible to manage access control using FC zoning
Vendor specific implementations should extend this class to provide
concrete implementation for add_connection and delete_connection
interfaces.

**Related Flags**

:zoning_policy: Used by: class: 'FCZoneDriver'. Defaults to 'none'
:zone_driver: Used by: class: 'FCZoneDriver'. Defaults to 'none'

"""

from oslo_log import log as logging

from cinder.interface import fczm_driver
from cinder.zonemanager import fc_common

LOG = logging.getLogger(__name__)


class FCZoneDriver(
        fc_common.FCCommon, fczm_driver.FibreChannelZoneManagerDriver):
    """Interface to manage Connection control during attach/detach."""

    # If a driver hasn't maintained their CI system, this will get set
    # to False, which prevents the driver from starting.
    # Add enable_unsupported_driver = True in cinder.conf to get the
    # unsupported driver started.
    SUPPORTED = True

    def __init__(self, **kwargs):
        super(FCZoneDriver, self).__init__(**kwargs)
        LOG.debug("Initializing FCZoneDriver")

    @property
    def supported(self):
        return self.SUPPORTED



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest
from unittest import mock

from oslo_utils.secretutils import md5

from cinder import exception
from cinder.tests.unit import test
from cinder.volume import configuration as conf
from cinder.volume.drivers.kioxia import entities
from cinder.volume.drivers.kioxia import kumoscale as kioxia
from cinder.volume.drivers.kioxia import rest_client

VOL_BACKEND_NAME = 'kioxia_kumoscale_1'
VOL_NAME = 'volume-c2fd04e3-320e-44eb-b-2'
VOL_UUID = 'c20aba21-6ef6-446b-b374-45733b4883ba'
VOL_SIZE = 10
VOL_PROTOCOL = 'NVMeoF'
SNAP_UUID = 'c9ef9d49-0d26-44cb-b609-0b8bd2d3db77'
CONN_UUID = '34206309-3733-4cc6-a7d5-9d4dbbe377da'
CONN_HOST_NAME = 'devstack'
CONN_NQN = 'nqn.2014-08.org.nvmexpress:uuid:' \
           'beaae2de-3a97-4be1-a739-6ac4bc5bf138'
success_prov_response = entities.ProvisionerResponse(None, None, "Success",
                                                     "Success")
fail_prov_response = entities.ProvisionerResponse(None, None, "Failure",
                                                  "Failure")
prov_backend1 = entities.Backend(None, None, None, None, 'dummy-pid-1')
prov_backend2 = entities.Backend(None, None, None, None, 'dummy-pid-2')
prov_location1 = entities.Location(VOL_UUID, prov_backend1)
prov_location2 = entities.Location(VOL_UUID, prov_backend2)
prov_volume = entities.VolumeProv(VOL_UUID, None, None, None,
                                  None, None, None, None, None, None,
                                  None, True, None, [prov_location1,
                                                     prov_location2])
prov_volumes_response = entities.ProvisionerResponse([prov_volume])
no_entities_prov_response = entities.ProvisionerResponse([], None, "Success")


class KioxiaVolumeTestCase(test.TestCase):
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_info')
    @mock.patch.object(kioxia.KumoScaleBaseVolumeDriver, '_get_kumoscale')
    def setUp(self, mock_kumoscale, mock_get_info):
        mock_get_info.return_value = success_prov_response
        mock_kumoscale.return_value = \
            rest_client.KioxiaProvisioner(['1.2.3.4'], 'cert', 'token')
        super(KioxiaVolumeTestCase, self).setUp()
        self.cfg = mock.Mock(spec=conf.Configuration)
        self.cfg.volume_backend_name = VOL_BACKEND_NAME
        self.cfg.url = 'dummyURL'
        self.cfg.token = 'dummy.dummy.Rf-dummy-dummy-lE'
        self.cfg.cafile = 'dummy'
        self.cfg.num_replicas = 1
        self.cfg.block_size = 512
        self.cfg.max_iops_per_gb = 1000
        self.cfg.desired_iops_per_gb = 1000
        self.cfg.max_bw_per_gb = 1000
        self.cfg.desired_bw_per_gb = 1000
        self.cfg.same_rack_allowed = False
        self.cfg.max_replica_down_time = 5
        self.cfg.span_allowed = True
        self.cfg.vol_reserved_space_percentage = 20
        self.cfg.provisioning_type = 'THIN'
        self.driver = kioxia.KumoScaleBaseVolumeDriver(configuration=self.cfg)
        self.driver.configuration.get = lambda *args, **kwargs: {}
        self.driver.num_replicas = 2
        self.expected_stats = {
            'volume_backend_name': VOL_BACKEND_NAME,
            'vendor_name': 'KIOXIA',
            'driver_version': self.driver.VERSION,
            'storage_protocol': 'NVMeOF',
            'consistencygroup_support': False,
            'thin_provisioning_support': True,
            'multiattach': False,
            'total_capacity_gb': 1000,
            'free_capacity_gb': 600
        }

    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_info')
    def test_get_kumoscale(self, mock_get_info):
        mock_get_info.return_value = success_prov_response
        result = self.driver._get_kumoscale('https://1.2.3.4:8090', 'token',
                                            'cert')
        self.assertEqual(result.mgmt_ips, ['1.2.3.4'])
        self.assertEqual(result.port, '8090')
        self.assertEqual(result.token, 'token')

    @mock.patch.object(rest_client.KioxiaProvisioner, 'create_volume')
    def test_volume_create_success(self, mock_create_volume):
        testvol = _stub_volume()
        mock_create_volume.return_value = success_prov_response
        result = self.driver.create_volume(testvol)
        args, kwargs = mock_create_volume.call_args
        mock_call = args[0]
        self.assertEqual(mock_call.alias, testvol['name'][:27])
        self.assertEqual(mock_call.capacity, testvol['size'])
        self.assertEqual(mock_call.uuid, testvol['id'])
        self.assertEqual(mock_call.protocol, VOL_PROTOCOL)
        self.assertIsNone(result)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'create_volume')
    def test_volume_create_failure(self, mock_create_volume):
        testvol = _stub_volume()
        mock_create_volume.return_value = fail_prov_response
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_volume, testvol)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'create_volume')
    def test_volume_create_exception(self, mock_create_volume):
        testvol = _stub_volume()
        mock_create_volume.side_effect = Exception()
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_volume, testvol)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'delete_volume')
    def test_delete_volume_success(self, mock_delete_volume):
        testvol = _stub_volume()
        mock_delete_volume.return_value = success_prov_response
        result = self.driver.delete_volume(testvol)
        mock_delete_volume.assert_any_call(testvol['id'])
        self.assertIsNone(result)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'delete_volume')
    def test_delete_volume_failure(self, mock_delete_volume):
        testvol = _stub_volume()
        mock_delete_volume.return_value = fail_prov_response
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.delete_volume, testvol)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'delete_volume')
    def test_delete_volume_exception(self, mock_delete_volume):
        testvol = _stub_volume()
        mock_delete_volume.side_effect = Exception()
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.delete_volume, testvol)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_backend_by_id')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_targets')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_volumes_by_uuid')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'publish')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'host_probe')
    def test_initialize_connection(self, mock_host_probe,
                                   mock_publish,
                                   mock_get_volumes_by_uuid,
                                   mock_get_targets,
                                   mock_get_backend_by_id):
        testvol = _stub_volume()
        testconn = _stub_connector()
        prov_target1 = TargetEntity('target.nqn', prov_backend1)
        prov_portal = PortalEntity('1.2.3.4', 4420, 'TCP')
        backend = BackendEntity([prov_portal])
        prov_targets_response = entities.ProvisionerResponse([prov_target1])
        mock_publish.return_value = success_prov_response
        mock_host_probe.return_value = success_prov_response
        mock_get_volumes_by_uuid.return_value = prov_volumes_response
        mock_get_targets.return_value = prov_targets_response
        mock_get_backend_by_id.return_value = \
            entities.ProvisionerResponse([backend])
        result = self.driver.initialize_connection(testvol, testconn)
        mock_host_probe.assert_any_call(testconn['nqn'],
                                        testconn['uuid'],
                                        testconn['host'],
                                        'Agent', 'cinder-driver-0.1', 30)
        mock_publish.assert_any_call(testconn['uuid'], testvol['id'])
        mock_get_volumes_by_uuid.assert_any_call(testvol['id'])
        mock_get_targets.assert_any_call(testconn['uuid'], testvol['id'])
        mock_get_backend_by_id.assert_any_call('dummy-pid-1')
        expected_replica = {'portals': [('1.2.3.4', '4420', 'TCP')],
                            'target_nqn': 'target.nqn',
                            'vol_uuid': testvol['id']}
        expected_data = {
            'vol_uuid': testvol['id'],
            'alias': testvol['name'],
            'writable': True,
            'volume_replicas': [expected_replica]
        }
        expected_result = {
            'driver_volume_type': 'nvmeof',
            'data': expected_data
        }
        self.assertDictEqual(result, expected_result)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_backend_by_id')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_targets')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_volumes_by_uuid')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'publish')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'host_probe')
    def test_initialize_connection_host_probe_failure(self, mock_host_probe,
                                                      mock_publish,
                                                      mock_get_volumes_by_uuid,
                                                      mock_get_targets,
                                                      mock_get_backend_by_id):
        testvol = _stub_volume()
        testconn = _stub_connector()
        prov_target = TargetEntity('target.nqn', prov_backend1)
        prov_portal = PortalEntity('1.2.3.4', 4420, 'TCP')
        backend = BackendEntity([prov_portal])
        prov_targets_response = entities.ProvisionerResponse([prov_target])
        mock_publish.return_value = success_prov_response
        mock_host_probe.return_value = fail_prov_response
        mock_get_volumes_by_uuid.return_value = prov_volumes_response
        mock_get_targets.return_value = prov_targets_response
        mock_get_backend_by_id.return_value = \
            entities.ProvisionerResponse([backend])
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.initialize_connection, testvol, testconn)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_backend_by_id')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_targets')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_volumes_by_uuid')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'publish')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'host_probe')
    def test_initialize_connection_host_probe_exception(
            self, mock_host_probe, mock_publish, mock_get_volumes_by_uuid,
            mock_get_targets, mock_get_backend_by_id):
        testvol = _stub_volume()
        testconn = _stub_connector()
        prov_target = TargetEntity('target.nqn', prov_backend1)
        prov_portal = PortalEntity('1.2.3.4', 4420, 'TCP')
        backend = BackendEntity([prov_portal])
        prov_targets_response = entities.ProvisionerResponse([prov_target])
        mock_publish.return_value = success_prov_response
        mock_host_probe.side_effect = Exception()
        mock_get_volumes_by_uuid.return_value = prov_volumes_response
        mock_get_targets.return_value = prov_targets_response
        mock_get_backend_by_id.return_value = \
            entities.ProvisionerResponse([backend])
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.initialize_connection, testvol, testconn)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_backend_by_id')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_targets')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_volumes_by_uuid')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'publish')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'host_probe')
    def test_initialize_connection_publish_failure(self, mock_host_probe,
                                                   mock_publish,
                                                   mock_get_volumes_by_uuid,
                                                   mock_get_targets,
                                                   mock_get_backend_by_id):
        testvol = _stub_volume()
        testconn = _stub_connector()
        prov_target = TargetEntity('target.nqn', prov_backend1)
        prov_portal = PortalEntity('1.2.3.4', 4420, 'TCP')
        backend = BackendEntity([prov_portal])
        prov_targets_response = entities.ProvisionerResponse([prov_target])
        mock_publish.return_value = fail_prov_response
        mock_host_probe.return_value = success_prov_response
        mock_get_volumes_by_uuid.return_value = prov_volumes_response
        mock_get_targets.return_value = prov_targets_response
        mock_get_backend_by_id.return_value = \
            entities.ProvisionerResponse([backend])
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.initialize_connection, testvol, testconn)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_backend_by_id')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_targets')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_volumes_by_uuid')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'publish')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'host_probe')
    def test_initialize_connection_publish_exception(self, mock_host_probe,
                                                     mock_publish,
                                                     mock_get_volumes_by_uuid,
                                                     mock_get_targets,
                                                     mock_get_backend_by_id):
        testvol = _stub_volume()
        testconn = _stub_connector()
        prov_target = TargetEntity('target.nqn', prov_backend1)
        prov_portal = PortalEntity('1.2.3.4', 4420, 'TCP')
        backend = BackendEntity([prov_portal])
        prov_targets_response = entities.ProvisionerResponse([prov_target])
        mock_publish.side_effect = Exception()
        mock_host_probe.return_value = success_prov_response
        mock_get_volumes_by_uuid.return_value = prov_volumes_response
        mock_get_targets.return_value = prov_targets_response
        mock_get_backend_by_id.return_value = \
            entities.ProvisionerResponse([backend])
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.initialize_connection, testvol, testconn)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_backend_by_id')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_targets')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_volumes_by_uuid')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'publish')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'host_probe')
    def test_initialize_connection_volumes_failure(self, mock_host_probe,
                                                   mock_publish,
                                                   mock_get_volumes_by_uuid,
                                                   mock_get_targets,
                                                   mock_get_backend_by_id):
        testvol = _stub_volume()
        testconn = _stub_connector()
        prov_target = TargetEntity('target.nqn', prov_backend1)
        prov_portal = PortalEntity('1.2.3.4', 4420, 'TCP')
        backend = BackendEntity([prov_portal])
        prov_targets_response = entities.ProvisionerResponse([prov_target])
        mock_publish.return_value = success_prov_response
        mock_host_probe.return_value = success_prov_response
        mock_get_volumes_by_uuid.return_value = fail_prov_response
        mock_get_targets.return_value = prov_targets_response
        mock_get_backend_by_id.return_value = \
            entities.ProvisionerResponse([backend])
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.initialize_connection, testvol, testconn)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_backend_by_id')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_targets')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_volumes_by_uuid')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'publish')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'host_probe')
    def test_initialize_connection_no_volumes(self, mock_host_probe,
                                              mock_publish,
                                              mock_get_volumes_by_uuid,
                                              mock_get_targets,
                                              mock_get_backend_by_id):
        testvol = _stub_volume()
        testconn = _stub_connector()
        prov_target = TargetEntity('target.nqn', prov_backend1)
        prov_portal = PortalEntity('1.2.3.4', 4420, 'TCP')
        backend = BackendEntity([prov_portal])
        prov_targets_response = entities.ProvisionerResponse([prov_target])
        mock_publish.return_value = success_prov_response
        mock_host_probe.return_value = success_prov_response
        mock_get_volumes_by_uuid.return_value = no_entities_prov_response
        mock_get_targets.return_value = prov_targets_response
        mock_get_backend_by_id.return_value = \
            entities.ProvisionerResponse([backend])
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.initialize_connection, testvol, testconn)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_backend_by_id')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_targets')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_volumes_by_uuid')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'publish')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'host_probe')
    def test_initialize_connection_volumes_exception(self, mock_host_probe,
                                                     mock_publish,
                                                     mock_get_volumes_by_uuid,
                                                     mock_get_targets,
                                                     mock_get_backend_by_id):
        testvol = _stub_volume()
        testconn = _stub_connector()
        prov_target = TargetEntity('target.nqn', prov_backend1)
        prov_portal = PortalEntity('1.2.3.4', 4420, 'TCP')
        backend = BackendEntity([prov_portal])
        prov_targets_response = entities.ProvisionerResponse([prov_target])
        mock_publish.return_value = success_prov_response
        mock_host_probe.return_value = success_prov_response
        mock_get_volumes_by_uuid.side_effect = Exception()
        mock_get_targets.return_value = prov_targets_response
        mock_get_backend_by_id.return_value = \
            entities.ProvisionerResponse([backend])
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.initialize_connection, testvol, testconn)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_backend_by_id')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_targets')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_volumes_by_uuid')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'publish')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'host_probe')
    def test_initialize_connection_targets_failure(self, mock_host_probe,
                                                   mock_publish,
                                                   mock_get_volumes_by_uuid,
                                                   mock_get_targets,
                                                   mock_get_backend_by_id):
        testvol = _stub_volume()
        testconn = _stub_connector()
        prov_portal = PortalEntity('1.2.3.4', 4420, 'TCP')
        backend = BackendEntity([prov_portal])
        mock_publish.return_value = success_prov_response
        mock_host_probe.return_value = success_prov_response
        mock_get_volumes_by_uuid.return_value = prov_volumes_response
        mock_get_targets.return_value = fail_prov_response
        mock_get_backend_by_id.return_value = \
            entities.ProvisionerResponse([backend])
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.initialize_connection, testvol, testconn)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_backend_by_id')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_targets')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_volumes_by_uuid')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'publish')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'host_probe')
    def test_initialize_connection_no_targets(self, mock_host_probe,
                                              mock_publish,
                                              mock_get_volumes_by_uuid,
                                              mock_get_targets,
                                              mock_get_backend_by_id):
        testvol = _stub_volume()
        testconn = _stub_connector()
        prov_portal = PortalEntity('1.2.3.4', 4420, 'TCP')
        backend = BackendEntity([prov_portal])
        mock_publish.return_value = success_prov_response
        mock_host_probe.return_value = success_prov_response
        mock_get_volumes_by_uuid.return_value = prov_volumes_response
        mock_get_targets.return_value = no_entities_prov_response
        mock_get_backend_by_id.return_value = \
            entities.ProvisionerResponse([backend])
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.initialize_connection, testvol, testconn)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_backend_by_id')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_targets')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_volumes_by_uuid')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'publish')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'host_probe')
    def test_initialize_connection_targets_exception(self, mock_host_probe,
                                                     mock_publish,
                                                     mock_get_volumes_by_uuid,
                                                     mock_get_targets,
                                                     mock_get_backend_by_id):
        testvol = _stub_volume()
        testconn = _stub_connector()
        prov_portal = PortalEntity('1.2.3.4', 4420, 'TCP')
        backend = BackendEntity([prov_portal])
        mock_publish.return_value = success_prov_response
        mock_host_probe.return_value = success_prov_response
        mock_get_volumes_by_uuid.return_value = prov_volumes_response
        mock_get_targets.side_effect = Exception()
        mock_get_backend_by_id.return_value = \
            entities.ProvisionerResponse([backend])
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.initialize_connection, testvol, testconn)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_backend_by_id')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_targets')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_volumes_by_uuid')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'publish')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'host_probe')
    def test_initialize_connection_backend_failure(self, mock_host_probe,
                                                   mock_publish,
                                                   mock_get_volumes_by_uuid,
                                                   mock_get_targets,
                                                   mock_get_backend_by_id):
        testvol = _stub_volume()
        testconn = _stub_connector()
        prov_target = TargetEntity('target.nqn', prov_backend1)
        prov_targets_response = entities.ProvisionerResponse([prov_target])
        mock_publish.return_value = success_prov_response
        mock_host_probe.return_value = success_prov_response
        mock_get_volumes_by_uuid.return_value = prov_volumes_response
        mock_get_targets.return_value = prov_targets_response
        mock_get_backend_by_id.return_value = fail_prov_response
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.initialize_connection, testvol, testconn)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_backend_by_id')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_targets')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_volumes_by_uuid')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'publish')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'host_probe')
    def test_initialize_connection_no_backend(self, mock_host_probe,
                                              mock_publish,
                                              mock_get_volumes_by_uuid,
                                              mock_get_targets,
                                              mock_get_backend_by_id):
        testvol = _stub_volume()
        testconn = _stub_connector()
        prov_target = TargetEntity('target.nqn', prov_backend1)
        prov_targets_response = entities.ProvisionerResponse([prov_target])
        mock_publish.return_value = success_prov_response
        mock_host_probe.return_value = success_prov_response
        mock_get_volumes_by_uuid.return_value = prov_volumes_response
        mock_get_targets.return_value = prov_targets_response
        mock_get_backend_by_id.return_value = no_entities_prov_response
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.initialize_connection, testvol, testconn)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_backend_by_id')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_targets')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_volumes_by_uuid')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'publish')
    @mock.patch.object(rest_client.KioxiaProvisioner, 'host_probe')
    def test_initialize_connection_backend_exception(self, mock_host_probe,
                                                     mock_publish,
                                                     mock_get_volumes_by_uuid,
                                                     mock_get_targets,
                                                     mock_get_backend_by_id):
        testvol = _stub_volume()
        testconn = _stub_connector()
        prov_target = TargetEntity('target.nqn', prov_backend1)
        prov_targets_response = entities.ProvisionerResponse([prov_target])
        mock_publish.return_value = success_prov_response
        mock_host_probe.return_value = success_prov_response
        mock_get_volumes_by_uuid.return_value = prov_volumes_response
        mock_get_targets.return_value = prov_targets_response
        mock_get_backend_by_id.side_effect = Exception()
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.initialize_connection, testvol, testconn)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'unpublish')
    def test_terminate_connection(self, mock_unpublish):
        testvol = _stub_volume()
        testconn = _stub_connector()
        mock_unpublish.return_value = success_prov_response
        result = self.driver.terminate_connection(testvol, testconn)
        mock_unpublish.assert_any_call(testconn['uuid'], testvol['id'])
        self.assertIsNone(result)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'unpublish')
    def test_terminate_connection_unpublish_failure(self, mock_unpublish):
        testvol = _stub_volume()
        testconn = _stub_connector()
        mock_unpublish.return_value = fail_prov_response
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.terminate_connection, testvol, testconn)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'unpublish')
    def test_terminate_connection_unpublish_exception(self, mock_unpublish):
        testvol = _stub_volume()
        testconn = _stub_connector()
        mock_unpublish.side_effect = Exception()
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.terminate_connection, testvol, testconn)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_tenants')
    def test_get_volume_stats(self, mock_get_tenants):
        tenant = TenantEntity(1000, 400)
        mock_get_tenants.return_value = entities.ProvisionerResponse([tenant])
        result = self.driver.get_volume_stats(True)
        mock_get_tenants.assert_any_call()
        self.assertDictEqual(result, self.expected_stats)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_tenants')
    def test_get_volume_stats_tenants_failure(self, mock_get_tenants):
        mock_get_tenants.return_value = fail_prov_response
        self.expected_stats['total_capacity_gb'] = 'unknown'
        self.expected_stats['free_capacity_gb'] = 'unknown'
        self.assertDictEqual(
            self.driver.get_volume_stats(True), self.expected_stats)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_tenants')
    def test_get_volume_stats_no_tenants(self, mock_get_tenants):
        mock_get_tenants.return_value = no_entities_prov_response
        self.expected_stats['total_capacity_gb'] = 'unknown'
        self.expected_stats['free_capacity_gb'] = 'unknown'
        self.assertDictEqual(
            self.driver.get_volume_stats(True), self.expected_stats)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'get_tenants')
    def test_get_volume_stats_tenants_exception(self, mock_get_tenants):
        mock_get_tenants.side_effect = Exception()
        self.expected_stats['total_capacity_gb'] = 'unknown'
        self.expected_stats['free_capacity_gb'] = 'unknown'
        self.assertDictEqual(
            self.driver.get_volume_stats(True), self.expected_stats)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'create_snapshot')
    def test_create_snapshot_success(self, mock_create_snapshot):
        testsnap = _stub_snapshot()
        mock_create_snapshot.return_value = success_prov_response
        result = self.driver.create_snapshot(testsnap)
        args, kwargs = mock_create_snapshot.call_args
        mock_call = args[0]
        self.assertEqual(mock_call.alias, testsnap['name'])
        self.assertEqual(mock_call.volumeID, testsnap['volume_id'])
        self.assertEqual(mock_call.snapshotID, testsnap['id'])
        self.assertIsNone(result)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'create_snapshot')
    def test_create_snapshot_failure(self, mock_create_snapshot):
        testsnap = _stub_snapshot()
        mock_create_snapshot.return_value = fail_prov_response
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_snapshot, testsnap)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'create_snapshot')
    def test_create_snapshot_exception(self, mock_create_snapshot):
        testsnap = _stub_snapshot()
        mock_create_snapshot.side_effect = Exception()
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_snapshot, testsnap)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'delete_snapshot')
    def test_delete_snapshot_success(self, mock_delete_snapshot):
        testsnap = _stub_snapshot()
        mock_delete_snapshot.return_value = success_prov_response
        result = self.driver.delete_snapshot(testsnap)
        mock_delete_snapshot.assert_any_call(testsnap['id'])
        self.assertIsNone(result)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'delete_snapshot')
    def test_delete_snapshot_failure(self, mock_delete_snapshot):
        testsnap = _stub_snapshot()
        mock_delete_snapshot.return_value = fail_prov_response
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.delete_snapshot, testsnap)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'delete_snapshot')
    def test_delete_snapshot_exception(self, mock_delete_snapshot):
        testsnap = _stub_snapshot()
        mock_delete_snapshot.side_effect = Exception()
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.delete_snapshot, testsnap)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'create_snapshot_volume')
    def test_create_volume_from_snapshot_success(self,
                                                 mock_create_snapshot_volume):
        testsnap = _stub_snapshot()
        testvol = _stub_volume()
        mock_create_snapshot_volume.return_value = success_prov_response
        result = self.driver.create_volume_from_snapshot(testvol, testsnap)
        args, kwargs = mock_create_snapshot_volume.call_args
        mock_call = args[0]
        self.assertEqual(mock_call.alias, testvol['name'])
        self.assertEqual(mock_call.volumeID, testsnap['volume_id'])
        self.assertEqual(mock_call.snapshotID, testsnap['id'])
        self.assertEqual(mock_call.protocol, VOL_PROTOCOL)
        self.assertIsNone(result)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'create_snapshot_volume')
    def test_create_volume_from_snapshot_failure(self,
                                                 mock_create_snapshot_volume):
        testsnap = _stub_snapshot()
        testvol = _stub_volume()
        mock_create_snapshot_volume.return_value = fail_prov_response
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_volume_from_snapshot, testvol,
                          testsnap)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'create_snapshot_volume')
    def test_create_volume_from_snapshot_exception(
            self, mock_create_snapshot_volume):
        testsnap = _stub_snapshot()
        testvol = _stub_volume()
        mock_create_snapshot_volume.side_effect = Exception()
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_volume_from_snapshot, testvol,
                          testsnap)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'expand_volume')
    def test_extend_volume_success(self, mock_expand_volume):
        testvol = _stub_volume()
        mock_expand_volume.return_value = success_prov_response
        new_size = VOL_SIZE + 2
        result = self.driver.extend_volume(testvol, new_size)
        mock_expand_volume.assert_any_call(new_size, testvol['id'])
        self.assertIsNone(result)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'expand_volume')
    def test_extend_volume_failure(self, mock_expand_volume):
        testvol = _stub_volume()
        mock_expand_volume.return_value = fail_prov_response
        new_size = VOL_SIZE + 2
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.extend_volume, testvol, new_size)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'expand_volume')
    def test_extend_volume_exception(self, mock_expand_volume):
        testvol = _stub_volume()
        mock_expand_volume.side_effect = Exception()
        new_size = VOL_SIZE + 2
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.extend_volume, testvol, new_size)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'clone_volume')
    def test_create_cloned_volume_success(self, mock_clone_volume):
        testvol = _stub_volume()
        mock_clone_volume.return_value = success_prov_response
        result = self.driver.create_cloned_volume(testvol, testvol)
        args, kwargs = mock_clone_volume.call_args
        mock_call = args[0]
        self.assertEqual(mock_call.alias, testvol['name'])
        self.assertEqual(mock_call.capacity, testvol['size'])
        self.assertEqual(mock_call.volumeId, testvol['id'])
        self.assertEqual(mock_call.sourceVolumeId, testvol['id'])
        self.assertIsNone(result)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'clone_volume')
    def test_create_cloned_volume_failure(self, mock_clone_volume):
        testvol = _stub_volume()
        mock_clone_volume.return_value = fail_prov_response
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_cloned_volume, testvol, testvol)

    @mock.patch.object(rest_client.KioxiaProvisioner, 'clone_volume')
    def test_create_cloned_volume_exception(self, mock_clone_volume):
        testvol = _stub_volume()
        mock_clone_volume.side_effect = Exception()
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_cloned_volume, testvol, testvol)

    def test_convert_host_name(self):
        name = 'ks-node3-000c2960a794-000c2960a797'
        result = self.driver._convert_host_name(name)
        expected = md5(name.encode('utf-8'), usedforsecurity=False).hexdigest()
        self.assertEqual(result, expected)

    def test_create_export(self):
        result = self.driver.create_export(None, None, None)
        self.assertIsNone(result)

    def test_ensure_export(self):
        result = self.driver.ensure_export(None, None)
        self.assertIsNone(result)

    def test_remove_export(self):
        result = self.driver.remove_export(None, None)
        self.assertIsNone(result)

    def test_check_for_setup_error(self):
        result = self.driver.check_for_setup_error()
        self.assertIsNone(result)


def _stub_volume(*args, **kwargs):
    volume = {'id': kwargs.get('id', VOL_UUID),
              'name': kwargs.get('name', VOL_NAME),
              'project_id': "test-project",
              'display_name': kwargs.get('display_name', VOL_NAME),
              'size': kwargs.get('size', VOL_SIZE),
              'provider_location': kwargs.get('provider_location', None),
              'volume_type_id': kwargs.get('volume_type_id', None)}
    return volume


def _stub_connector(*args, **kwargs):
    connector = {'uuid': kwargs.get('uuid', CONN_UUID),
                 'nqn': kwargs.get('nqn', CONN_NQN),
                 'host': kwargs.get('host', CONN_HOST_NAME)}
    return connector


def _stub_snapshot(*args, **kwargs):
    volume = {'id': kwargs.get('id', SNAP_UUID),
              'name': kwargs.get('name', 'snap2000'),
              'volume_id': kwargs.get('id', VOL_UUID)}
    return volume


class TenantEntity:
    def __init__(self, capacity, consumed):
        self.tenantId = '0'
        self.capacity = capacity
        self.consumedCapacity = consumed


class TargetEntity:
    def __init__(self, name, backend):
        self.targetName = name
        self.backend = backend


class BackendEntity:
    def __init__(self, portals):
        self.portals = portals


class PortalEntity:
    def __init__(self, ip, port, transport):
        self.ip = ip
        self.port = port
        self.transport = transport


if __name__ == '__main__':
    unittest.main()



#   Copyright 2015 Huawei Technologies Co., Ltd.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
from http import HTTPStatus

from oslo_log import log as logging

from cinder.api.contrib import resource_common_manage
from cinder.api import extensions
from cinder.api.openstack import wsgi
from cinder.api.schemas import snapshot_manage
from cinder.api import validation
from cinder.api.views import manageable_snapshots as list_manageable_view
from cinder.api.views import snapshots as snapshot_views
from cinder.policies import manageable_snapshots as policy
from cinder import volume as cinder_volume

LOG = logging.getLogger(__name__)


class SnapshotManageController(wsgi.Controller):
    """The /os-snapshot-manage controller for the OpenStack API."""

    _view_builder_class = snapshot_views.ViewBuilder

    def __init__(self, *args, **kwargs):
        super(SnapshotManageController, self).__init__(*args, **kwargs)
        self.volume_api = cinder_volume.API()
        self._list_manageable_view = list_manageable_view.ViewBuilder()

    @wsgi.response(HTTPStatus.ACCEPTED)
    @validation.schema(snapshot_manage.create)
    def create(self, req, body):
        """Instruct Cinder to manage a storage snapshot object.

        Manages an existing backend storage snapshot object (e.g. a Linux
        logical volume or a SAN disk) by creating the Cinder objects required
        to manage it, and possibly renaming the backend storage snapshot object
        (driver dependent).

        From an API perspective, this operation behaves very much like a
        snapshot creation operation.

        Required HTTP Body:

        .. code-block:: json

         {
           "snapshot":
           {
             "volume_id": "<Cinder volume already exists in volume backend>",
             "ref":
                "<Driver-specific reference to the existing storage object>"
           }
         }

        See the appropriate Cinder drivers' implementations of the
        manage_snapshot method to find out the accepted format of 'ref'.
        For example,in LVM driver, it will be the logic volume name of snapshot
        which you want to manage.

        This API call will return with an error if any of the above elements
        are missing from the request, or if the 'volume_id' element refers to
        a cinder volume that could not be found.

        The snapshot will later enter the error state if it is discovered that
        'ref' is bad.

        Optional elements to 'snapshot' are::

         name           A name for the new snapshot.
         description    A description for the new snapshot.
         metadata       Key/value pairs to be associated with the new snapshot.

        """
        context = req.environ['cinder.context']

        snapshot = body['snapshot']
        # Check whether volume exists
        volume_id = snapshot['volume_id']
        # Not found exception will be handled at the wsgi level
        volume = self.volume_api.get(context, volume_id)
        context.authorize(policy.MANAGE_POLICY, target_obj=volume)

        LOG.debug('Manage snapshot request body: %s', body)

        snapshot_parameters = {}

        snapshot_parameters['metadata'] = snapshot.get('metadata', None)
        snapshot_parameters['description'] = snapshot.get('description', None)
        snapshot_parameters['name'] = snapshot.get('name')

        # Not found exception will be handled at the wsgi level
        new_snapshot = self.volume_api.manage_existing_snapshot(
            context,
            snapshot['ref'],
            volume,
            **snapshot_parameters)

        return self._view_builder.detail(req, new_snapshot)

    @wsgi.extends
    def index(self, req):
        """Returns a summary list of snapshots available to manage."""
        context = req.environ['cinder.context']
        context.authorize(policy.LIST_MANAGEABLE_POLICY)
        return resource_common_manage.get_manageable_resources(
            req, False, self.volume_api.get_manageable_snapshots,
            self._list_manageable_view)

    @wsgi.extends
    def detail(self, req):
        """Returns a detailed list of snapshots available to manage."""
        context = req.environ['cinder.context']
        context.authorize(policy.LIST_MANAGEABLE_POLICY)
        return resource_common_manage.get_manageable_resources(
            req, True, self.volume_api.get_manageable_snapshots,
            self._list_manageable_view)


class Snapshot_manage(extensions.ExtensionDescriptor):
    """Allows existing backend storage to be 'managed' by Cinder."""

    name = 'SnapshotManage'
    alias = 'os-snapshot-manage'
    updated = '2014-12-31T00:00:00+00:00'

    def get_resources(self):
        controller = SnapshotManageController()
        return [extensions.ResourceExtension(Snapshot_manage.alias,
                                             controller,
                                             collection_actions=
                                             {'detail': 'GET'})]



# Copyright (c) 2016 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from http import HTTPStatus

from oslo_utils import strutils
from oslo_utils import timeutils

from cinder.api import microversions as mv
from cinder.api.openstack import wsgi
from cinder.api.schemas import workers
from cinder.api.v3.views import workers as workers_view
from cinder.api import validation
from cinder import db
from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder.objects import cleanable
from cinder.policies import workers as policy
from cinder.scheduler import rpcapi as sch_rpc
from cinder import utils


class WorkerController(wsgi.Controller):

    def __init__(self, *args, **kwargs):
        self.sch_api = sch_rpc.SchedulerAPI()

    @wsgi.Controller.api_version(mv.WORKERS_CLEANUP)
    @wsgi.response(HTTPStatus.ACCEPTED)
    @validation.schema(workers.cleanup)
    def cleanup(self, req, body=None):
        """Do the cleanup on resources from a specific service/host/node."""
        # Let the wsgi middleware convert NotAuthorized exceptions
        ctxt = req.environ['cinder.context']
        ctxt.authorize(policy.CLEAN_POLICY)
        body = body or {}

        for boolean in ('disabled', 'is_up'):
            if body.get(boolean) is not None:
                body[boolean] = strutils.bool_from_string(body[boolean])

        resource_type = body.get('resource_type')

        if resource_type:
            resource_type = resource_type.title()
            types = cleanable.CinderCleanableObject.cleanable_resource_types
            if resource_type not in types:
                valid_types = utils.build_or_str(types)
                msg = _('Resource type %(resource_type)s not valid,'
                        ' must be %(valid_types)s')
                msg = msg % {"resource_type": resource_type,
                             "valid_types": valid_types}
                raise exception.InvalidInput(reason=msg)
            body['resource_type'] = resource_type

        resource_id = body.get('resource_id')
        if resource_id:

            # If we have the resource type but we don't have where it is
            # located, we get it from the DB to limit the distribution of the
            # request by the scheduler, otherwise it will be distributed to all
            # the services.
            location_keys = {'service_id', 'cluster_name', 'host'}
            if not location_keys.intersection(body):
                workers = db.worker_get_all(ctxt, resource_id=resource_id,
                                            binary=body.get('binary'),
                                            resource_type=resource_type)

                if len(workers) == 0:
                    msg = (_('There is no resource with UUID %s pending '
                             'cleanup.'), resource_id)
                    raise exception.InvalidInput(reason=msg)
                if len(workers) > 1:
                    msg = (_('There are multiple resources with UUID %s '
                             'pending cleanup.  Please be more specific.'),
                           resource_id)
                    raise exception.InvalidInput(reason=msg)

                worker = workers[0]
                body.update(service_id=worker.service_id,
                            resource_type=worker.resource_type)

        body['until'] = timeutils.utcnow()

        # NOTE(geguileo): If is_up is not specified in the request
        # CleanupRequest's default will be used (False)
        cleanup_request = objects.CleanupRequest(**body)
        cleaning, unavailable = self.sch_api.work_cleanup(ctxt,
                                                          cleanup_request)
        return {
            'cleaning': workers_view.ViewBuilder.service_list(cleaning),
            'unavailable': workers_view.ViewBuilder.service_list(unavailable),
        }


def create_resource():
    return wsgi.Resource(WorkerController())



# Copyright 2021 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from migrate.changeset import constraint
import sqlalchemy as sa


def upgrade(migrate_engine):
    """Update quota_usages table to prevent races on creation.

    Add race_preventer field and a unique constraint to prevent quota usage
    duplicates and races that mess the quota system when first creating rows.
    """
    # There's no need to set the race_preventer field for existing DB entries,
    # since the race we want to prevent is only on creation.
    meta = sa.MetaData(bind=migrate_engine)
    quota_usages = sa.Table('quota_usages', meta, autoload=True)

    if not hasattr(quota_usages.c, 'race_preventer'):
        quota_usages.create_column(
            sa.Column('race_preventer', sa.Boolean, nullable=True))

    unique = constraint.UniqueConstraint(
        'project_id', 'resource', 'race_preventer',
        table=quota_usages)
    unique.create(engine=migrate_engine)



#  Copyright (c) 2016 IBM Corporation
#  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import base64


def encrypt(string):
    return base64.b64encode(string.encode('UTF-8'))


def decrypt(string):
    missing_padding = len(string) % 4
    if missing_padding != 0:
        string += b'=' * (4 - missing_padding)
    return base64.b64decode(string)



#    Copyright 2015 SimpliVity Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt
from sqlalchemy.orm import attributes

from cinder import db
from cinder import objects
from cinder.objects import fields
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_volume
from cinder.tests.unit import objects as test_objects


@ddt.ddt
class TestVolumeAttachment(test_objects.BaseObjectsTestCase):

    @mock.patch('cinder.db.sqlalchemy.api.volume_attachment_get')
    def test_get_by_id(self, volume_attachment_get):
        db_attachment = fake_volume.volume_attachment_db_obj()
        attachment_obj = fake_volume.volume_attachment_ovo(self.context)
        volume_attachment_get.return_value = db_attachment
        attachment = objects.VolumeAttachment.get_by_id(self.context,
                                                        fake.ATTACHMENT_ID)
        self._compare(self, attachment_obj, attachment)

    @mock.patch.object(objects.Volume, 'get_by_id')
    def test_lazy_load_volume(self, volume_get_mock):
        volume = objects.Volume(self.context, id=fake.VOLUME_ID)
        volume_get_mock.return_value = volume
        attach = objects.VolumeAttachment(self.context, id=fake.ATTACHMENT_ID,
                                          volume_id=volume.id)

        r = attach.volume
        self.assertEqual(volume, r)
        volume_get_mock.assert_called_once_with(self.context, volume.id)

    def test_from_db_object_no_volume(self):
        original_get = attributes.InstrumentedAttribute.__get__

        def my_get(get_self, instance, owner):
            self.assertNotEqual('volume', get_self.key)
            return original_get(get_self, instance, owner)

        # Volume field is not loaded
        attach = fake_volume.models.VolumeAttachment(id=fake.ATTACHMENT_ID,
                                                     volume_id=fake.VOLUME_ID)
        patch_str = 'sqlalchemy.orm.attributes.InstrumentedAttribute.__get__'
        with mock.patch(patch_str, side_effect=my_get):
            objects.VolumeAttachment._from_db_object(
                self.context, objects.VolumeAttachment(), attach)

    @mock.patch('cinder.db.volume_attachment_update')
    def test_save(self, volume_attachment_update):
        attachment = fake_volume.volume_attachment_ovo(self.context)
        attachment.attach_status = fields.VolumeAttachStatus.ATTACHING
        attachment.save()
        volume_attachment_update.assert_called_once_with(
            self.context, attachment.id,
            {'attach_status': fields.VolumeAttachStatus.ATTACHING})

    @mock.patch('cinder.db.sqlalchemy.api.volume_attachment_get')
    def test_refresh(self, attachment_get):
        db_attachment1 = fake_volume.volume_attachment_db_obj()
        attachment_obj1 = fake_volume.volume_attachment_ovo(self.context)
        db_attachment2 = fake_volume.volume_attachment_db_obj()
        db_attachment2.mountpoint = '/dev/sdc'
        attachment_obj2 = fake_volume.volume_attachment_ovo(
            self.context, mountpoint='/dev/sdc')

        # On the second volume_attachment_get, return the volume attachment
        # with an updated mountpoint
        attachment_get.side_effect = [db_attachment1, db_attachment2]
        attachment = objects.VolumeAttachment.get_by_id(self.context,
                                                        fake.ATTACHMENT_ID)
        self._compare(self, attachment_obj1, attachment)

        # mountpoint was updated, so a volume attachment refresh should have a
        # new value for that field
        attachment.refresh()
        self._compare(self, attachment_obj2, attachment)
        attachment_get.assert_has_calls([mock.call(self.context,
                                                   fake.ATTACHMENT_ID),
                                         mock.call.__bool__(),
                                         mock.call(self.context,
                                                   fake.ATTACHMENT_ID)])

    @mock.patch('cinder.db.sqlalchemy.api.volume_attached')
    def test_volume_attached(self, volume_attached):
        attachment = fake_volume.volume_attachment_ovo(self.context)
        updated_values = {'mountpoint': '/dev/sda',
                          'attach_status': fields.VolumeAttachStatus.ATTACHED,
                          'instance_uuid': fake.INSTANCE_ID}
        volume_attached.return_value = (fake_volume.fake_db_volume(),
                                        updated_values)
        volume = attachment.finish_attach(fake.INSTANCE_ID,
                                          'fake_host',
                                          '/dev/sda',
                                          'rw')
        self.assertIsInstance(volume, objects.Volume)
        volume_attached.assert_called_once_with(mock.ANY,
                                                attachment.id,
                                                fake.INSTANCE_ID,
                                                'fake_host',
                                                '/dev/sda',
                                                'rw',
                                                True)
        self.assertEqual('/dev/sda', attachment.mountpoint)
        self.assertEqual(fake.INSTANCE_ID, attachment.instance_uuid)
        self.assertEqual(fields.VolumeAttachStatus.ATTACHED,
                         attachment.attach_status)

    def test_migrate_attachment_specs(self):
        # Create an attachment.
        attachment = objects.VolumeAttachment(
            self.context, attach_status='attaching', volume_id=fake.VOLUME_ID)
        attachment.create()
        # Create some attachment_specs. Note that the key and value have to
        # be strings, the table doesn't handle things like a wwpns list
        # for a fibrechannel connector.
        connector = {'host': '127.0.0.1'}
        db.attachment_specs_update_or_create(
            self.context, attachment.id, connector)
        # Now get the volume attachment object from the database and make
        # sure the connector was migrated from the attachment_specs table
        # to the volume_attachment table and the specs were deleted.
        attachment = objects.VolumeAttachment.get_by_id(
            self.context, attachment.id)
        self.assertIn('connector', attachment)
        self.assertDictEqual(connector, attachment.connector)
        self.assertEqual(0, len(db.attachment_specs_get(
            self.context, attachment.id)))
        # Make sure we can store a fibrechannel type connector that has a wwpns
        # list value.
        connector['wwpns'] = ['21000024ff34c92d', '21000024ff34c92c']
        attachment.connector = connector
        attachment.save()
        # Get the object from the DB again and make sure the connector is
        # there.
        attachment = objects.VolumeAttachment.get_by_id(
            self.context, attachment.id)
        self.assertIn('connector', attachment)
        self.assertDictEqual(connector, attachment.connector)


class TestVolumeAttachmentList(test_objects.BaseObjectsTestCase):
    @mock.patch('cinder.db.volume_attachment_get_all_by_volume_id')
    def test_get_all_by_volume_id(self, get_used_by_volume_id):
        db_attachment = fake_volume.volume_attachment_db_obj()
        get_used_by_volume_id.return_value = [db_attachment]
        attachment_obj = fake_volume.volume_attachment_ovo(self.context)

        attachments = objects.VolumeAttachmentList.get_all_by_volume_id(
            self.context, mock.sentinel.volume_id)

        self.assertEqual(1, len(attachments))
        self._compare(self, attachment_obj, attachments[0])

    @mock.patch('cinder.db.volume_attachment_get_all_by_host')
    def test_get_all_by_host(self, get_by_host):
        db_attachment = fake_volume.volume_attachment_db_obj()
        attachment_obj = fake_volume.volume_attachment_ovo(self.context)
        get_by_host.return_value = [db_attachment]

        attachments = objects.VolumeAttachmentList.get_all_by_host(
            self.context, mock.sentinel.host)
        self.assertEqual(1, len(attachments))
        self._compare(self, attachment_obj, attachments[0])

    @mock.patch('cinder.db.volume_attachment_get_all_by_instance_uuid')
    def test_get_all_by_instance_uuid(self, get_by_instance_uuid):
        db_attachment = fake_volume.volume_attachment_db_obj()
        get_by_instance_uuid.return_value = [db_attachment]
        attachment_obj = fake_volume.volume_attachment_ovo(self.context)

        attachments = objects.VolumeAttachmentList.get_all_by_instance_uuid(
            self.context, mock.sentinel.uuid)
        self.assertEqual(1, len(attachments))
        self._compare(self, attachment_obj, attachments[0])
